import moai.nn.utils as miu

import torch
import logging

log = logging.getLogger(__name__)

__all__ = [
    "make_residual_block"
]

import moai.nn.residual.standard as mistd
import moai.nn.residual.bottleneck as mibtl

__BLOCK_FACTORY__ = {# (b), (c) and (e) from https://towardsdatascience.com/an-overview-of-resnet-and-its-variants-5281e2f56035
    "standard":                                     mistd.Standard,                       # (b): y    =   A2(x  +   W2     *   A1(W1   *   x))
    "preresidual":                                  mistd.PreResidual,                    # (c): y    =   x    +   A2(W2   *   A1(W1   *   w))
    "preactivation":                                mistd.PreActivation,                  # (e): y    =   x    +   W2      *   A2(W1   *   A1(x))
    # bottleneck versions
    "bottleneck":                                   mibtl.Bottleneck,                     # (b): y    =   A3(x    +   W3  *   A2(W2   *   A1(W1   *   x)))
    "preres_bottleneck":                            mibtl.PreResBottleneck,               # (c): y    =   x       +  A3(W3    *   A2(W2   *   A1(W1   *   x)))
    "preactiv_bottleneck":                          mibtl.PreActivBottleneck,             # (e): y    =   x       +  W3   *   A3(W2   *   A2(W1   *   A1(x)))
}

del mistd
del mibtl
        
def make_residual_block(
    block_type: str,
    convolution_type: str,
    in_features: int,
    out_features: int,
    bottleneck_features: int,
    activation_type: str,
    strided: bool,
    convolution_params: dict={},
    downscale_params: dict={},
    activation_params: dict={"inplace": True},
    **kwargs
) -> torch.nn.Module:        
    if block_type in __BLOCK_FACTORY__.keys():
        return miu.instantiate(__BLOCK_FACTORY__[block_type],
            **{ 
                **locals(),
                "convolution_params": convolution_params,
                "downscale_params": downscale_params,
                "activation_params": activation_params,
                **kwargs
            })
    else:
        log.error("Residual block type (%s) not found." % block_type)


import moai.nn.activation as mia
import moai.nn.linear as miln

import torch
import functools
import logging

log = logging.getLogger(__name__)

__all__ = [
    "LinearBlock",
]

class LinearBlock(torch.nn.Module):
    def __init__(self,
        linear_type:str,
        activation_type: str,
        in_features: int,
        out_features: int,
        linear_params: dict,
        activation_params: dict,
        **kwargs: dict
    ):
        super(LinearBlock, self).__init__()
        self.linear = miln.make_linear_op(
            linear_type=linear_type,
            in_features=in_features,
            out_features=out_features,
            **linear_params #TODO: either merge kwargs here or in the factory method
        )
        self.activation = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )
        self.in_features = in_features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not hasattr(self, "view"):
            s = x.size()
            def conv_view(x: torch.Tensor, f:int) -> torch.Tensor:
                return x.view(-1, f)
            self.view = functools.partial(conv_view, f=self.in_features)\
                if len(s) > 2 else torch.nn.Identity()
        return self.activation(self.linear(self.view(x)))


from collections.abc import Callable

import torch
import functools
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["PredictionBias"]

class PredictionBias(Callable):
    def __init__(self, 
        bias:           float=1.0,
        kernel_size:    int=3,
        out_channels:   int=1,
    ):
        self.bias = bias
        self.out_channels = out_channels
        self.size = (kernel_size, kernel_size)

    def __call__(self,
        module: torch.nn.Module
    ) -> None:        
        if isinstance(module, torch.nn.Conv2d):#TODO: other conv types?
            if module.out_channels == self.out_channels and module.kernel_size == self.size:
                for k, v in module.named_parameters():
                    if k == 'bias':
                        log.info(f"Bias initialization for {module} to {self.bias}.")
                        v.data.fill_(self.bias)



from ts.torch_handler.base_handler import BaseHandler
from hydra.experimental import (
    compose,
    initialize,
)
from omegaconf.omegaconf import OmegaConf

import hydra.utils as hyu
import toolz
import os
import torch
import zipfile
import yaml
import logging
import typing

log = logging.getLogger(__name__)

__all__ = ['ModelServer']

class ModelServer(BaseHandler):
    def __init__(self) -> None:
        super().__init__()

    def _extract_files(self):
        with zipfile.ZipFile('conf.zip', 'r') as conf_zip:
            conf_zip.extractall('.')
        with zipfile.ZipFile('src.zip', 'r') as src_zip:
            src_zip.extractall('.')

    def _get_overrides(self):
        if os.path.exists('overrides.yaml'):
            with open('overrides.yaml') as f:
                return yaml.load(f, Loader=yaml.FullLoader)
        else:
            return []

    def initialize(self, context):        
        properties = context.system_properties
        self.map_location = "cuda" if torch.cuda.is_available(
        ) and properties.get("gpu_id") is not None else "cpu"
        self.device = torch.device(
            self.map_location + ":" + str(properties.get("gpu_id"))
            if torch.cuda.is_available() and properties.get("gpu_id") is not None
            else self.map_location
        )
        #NOTE: IMPORTANT!!!! DEBUG WHILE TRAINING ONLY !!!
        self.device = torch.device('cpu')
        log.warning("IMPORTANT: Model explicitly set to CPU mode for debugging purposes!")
        #NOTE: IMPORTANT!!!! DEBUG WHILE TRAINING ONLY !!!
        main_conf = context.manifest['model']['modelName'].replace('_', '/')
        log.info(f"Loading the {main_conf} endpoint.")
        self._extract_files()        
        try:
            with initialize(config_path="conf", job_name=main_conf):
                cfg = compose(
                    config_name=main_conf, 
                    overrides=self._get_overrides()
                )
                self.model = hyu.instantiate(cfg.model)
        except Exception as e:
            log.error(f"An error has occured while loading the model:\n{e}")
        self.model = self.model.to(self.device)
        self.model.eval()
        self.initialized = True
        log.info(f"Model ({type(self.model.inner if hasattr(self.model, 'inner') else self.model)}) loaded successfully.")        
        try:
            handler_overrides = None
            if os.path.exists('handler_overrides.yaml'):
                with open('handler_overrides.yaml') as f:
                    handler_overrides = yaml.load(f, Loader=yaml.FullLoader)
                    log.debug(f"Loaded handler overrides:\n{handler_overrides}")
            with initialize(config_path="conf", job_name=f"{main_conf}_preprocess_handlers"):
                cfg = compose(config_name="../pre")
                if handler_overrides is not None and 'preprocess' in handler_overrides.get('handlers', {}):
                    cfg = OmegaConf.merge(cfg, {'handlers': {'preprocess': handler_overrides['handlers']['preprocess']}})
                    log.debug(f"Merged handler overrides:\n{cfg}")
                self.preproc = {k: hyu.instantiate(h) for k, h in (toolz.get_in(['handlers', 'preprocess'], cfg) or {}).items()}
            with initialize(config_path="conf", job_name=f"{main_conf}_postprocess_handlers"):
                cfg = compose(config_name="../post")
                if handler_overrides is not None and 'postprocess' in handler_overrides.get('handlers', {}):
                    cfg = OmegaConf.merge(cfg, {'handlers': {'postprocess': handler_overrides['handlers']['postprocess']}})
                    log.debug(f"Merged handler overrides:\n{cfg}")
                self.postproc = {k: hyu.instantiate(h) for k, h in (toolz.get_in(['handlers', 'postprocess'], cfg) or {}).items()}
        except Exception as e:
            log.error(f"An error has occured while loading the handlers:\n{e}")
        self.model.initialize_parameters()

    def preprocess(self, 
        data:   typing.Mapping[str, typing.Any],
    ) -> typing.Dict[str, torch.Tensor]:
        log.debug(f"Preprocessing input:\n{data}")
        tensors = { '__moai__': { 'json': data } }
        #TODO: need to check batched inference
        body = data[0].get('body') or data[0].get('raw')
        for k, p in self.preproc.items():
            tensors = toolz.merge(tensors, p(body, self.device))
        log.debug(f"Tensors: {tensors.keys()}")
        return tensors
    
    def inference(self, 
        data:       typing.Mapping[str, torch.Tensor],
    ):
        metrics, tensors = self.model.test_step(data, batch_nb=0)
        return tensors
       
    def postprocess(self,
        data: typing.Mapping[str, torch.Tensor]
    ) -> typing.Sequence[typing.Any]:
        outs = []
        for k, p in self.postproc.items():
            outs.append(p(data, data['__moai__']['json']))
        return outs

    # def handle(self, data, context):
    #     return super(data, context)


from moai.engine.modules.seed import RandomSeed

import hydra.utils as hyu
import omegaconf.omegaconf

__all__ = [
    "Engine",  
]

class Engine(object):
    def __init__(self,
        modules: omegaconf.DictConfig=None
    ):        
        self.inits = [hyu.instantiate(conf) for conf in modules.values()]\
            if modules is not None else []
        if modules is None or 'manual_seed' not in modules.keys():
            RandomSeed()


import hydra
import sys
import omegaconf.omegaconf
import logging
import typing

from moai.engine.callbacks import ModelCallbacks

log = logging.getLogger(__name__)

def assign(cfg: omegaconf.DictConfig, attr: str) -> typing.Union[typing.Any]:
    return getattr(cfg, attr) if hasattr(cfg, attr) else None

def fit(cfg):
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    with open("config_resolved.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=True, sort_keys=False))
    with open("config.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=False, sort_keys=False))
    engine = hydra.utils.instantiate(cfg.engine)
    model = hydra.utils.instantiate(cfg.model,
        data=cfg.data,
        visualization=assign(cfg, "visualization"),
        export=assign(cfg, "export"),    
    )
    for name, remodel in (assign(cfg ,"remodel") or {}).items():
        hydra.utils.instantiate(remodel)(model)
    fitter = hydra.utils.instantiate(cfg.fitter, 
        logging=assign(cfg, "logging"),
        model_callbacks=ModelCallbacks(model=model),
    )
    log.info("Fitting started.")     
    fitter.run(model)
    log.info("Fitting completed.")

if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    config_filename = sys.argv.pop(1) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}")    
    fit = hydra.main(config_path="conf", config_name=config_filename)(fit)
    fit()


from moai.data.datasets.common import load_npz_files
from moai.utils.arguments import ensure_string_list

import torch
import glob
import os
import typing
import logging
import toolz

__all__ = ["Npz"]

log = logging.getLogger(__name__)

class Npz(torch.utils.data.Dataset):            #TODO add the loading different npz and combining them in the same dict case
    def __init__(self,
        root:           str='',
        **kwargs:       typing.Mapping[str, typing.Mapping[str, typing.Any]],
    ):
        self.files = glob.glob(os.path.join(root,'*.npz'))

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        ret = load_npz_files(self.files[index])
        return ret


import moai.utils.engine as mieng

import torch
import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ["Initializer"]

class Initializer(mieng.Single):
    def __init__(self, 
        schemes: omegaconf.DictConfig
    ):
        super(Initializer, self).__init__(items=schemes, name="scheme")

    def __call__(self, 
        model: torch.nn.Module
    ) -> None:
        log.info(f"Applying {len(self.items())} parameter initialization schemes:")
        for scheme in self.items():
            log.info(f"\tApplying {scheme.__class__.__name__}")
            model.apply(scheme)


from moai.utils.color.colormap import colormap

import numpy
import visdom
import pytorch_lightning
import logging
import typing
import toolz
from json2html import *

log = logging.getLogger(__name__)

__all__ = ["Visdom"]

class Visdom(pytorch_lightning.loggers.base.LightningLoggerBase):
    def __init__(self,
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,
        version:        int=0,
        train_key:      str="train_loss",
        test_key:       str="test_results",
        val_key:        str="val_metrics",
        plots_key:      str="plot",
        epoch_key:      str="epoch",
        step_key:       str="global_step",
        clear_window:   bool=True
    ):
        super(Visdom, self).__init__()
        # lighting logger
        self._rank, self._version = 0, version
        # visdom
        self.env_name, self.ip, self.port, self.viz = name, ip, port, None 
        # keys  
        self.keys: typing.Dict[str, str]={ }
        self.keys.update({
            'train': train_key, 'val': val_key, 'test': test_key,
            'loss_plots': plots_key, 'epoch': epoch_key, 'step': step_key
        })
        # plotting
        self.colors = colormap().astype(numpy.int32)
        # other
        self.colors: numpy.array
        self.best_epoch_loss: typing.List[float]=[]
        self.best_epoch_step: typing.List[int]=[]
        self.best_epoch_val_loss: typing.Dict[str, typing.List[float]]={ }
        self.best_epoch_val_step: typing.Dict[str, typing.List[int]]={ }
        self.loss_plots: typing.Dict[str, str]={ }
        self.metric_plots: typing.Dict[str, str]={ }
        self.clear_window = clear_window

    @property
    def name(self) -> str:
        return self.env_name

    @property
    def visualizer(self) -> visdom.Visdom:
        if self.viz is not None:
            return self.viz

        self.viz = visdom.Visdom(
            server=self.ip,
            port=self.port,
            env=self.name,
            use_incoming_socket=False
        )
        return self.viz

    def _plot_train_loss(self,
        epoch: int,
        value: float,
        step: int
    ) -> None:
        if not self.keys['train'] in self.loss_plots.keys():
            self.loss_plots[self.keys['train']] = self.visualizer.line(
                X=numpy.array([step,step]),
                Y=numpy.array([value,value]),
                env=self.name,
                name=self.keys['train'],
                opts={
                    'linecolor': numpy.array([self.colors[0]]), # first color
                    # 'legend': [str(self.keys['train']), str(self.keys['epoch'])],
                    'legend': [str(self.keys['train']), 'train_' + str(self.keys['epoch'])],
                    'title': str(self.keys['train']),
                    'xlabel':'steps',
                    'ylabel': str(self.keys['train'])
                }
            )
        else:
            self.visualizer.line(
                X=numpy.array([step]),
                Y=numpy.array([value]),
                win=self.loss_plots[self.keys['train']],
                name=self.keys['train'], 
                update='append',
            )
        while len(self.best_epoch_loss) <= epoch:
            self.best_epoch_loss.append(numpy.Infinity)
            self.best_epoch_step.append(step)        
        if value < self.best_epoch_loss[epoch]:
            self.best_epoch_loss[epoch] = value
            self.best_epoch_step[epoch] = step
            self.visualizer.line(
                X=numpy.array(self.best_epoch_step),
                Y=numpy.array(self.best_epoch_loss),
                update='insert',
                win=self.loss_plots[self.keys['train']],
                name='train_' + self.keys['epoch'],
                opts={
                    'markers': True,
                    'linecolor': numpy.array([self.colors[-1]]) # last color
                }
            )

    def _plot_losses(self,
        losses:         typing.Dict[str, float],
        epoch:          int,
        step_num:       int,
    ) -> None:
        for i, (k, v) in enumerate(losses.items()):
            if not self.keys['loss_plots'] in self.loss_plots.keys():            
                self.loss_plots[self.keys['loss_plots']] = self.visualizer.line(
                    X=numpy.array([step_num,step_num]),
                    Y=numpy.array([v,v]),
                    env=self.name,
                    name=k,
                    opts={
                        'linecolor': numpy.array([self.colors[2 + i]]), # second color and on
                        'legend': [k for k in losses.keys()],
                        'title': self.keys['loss_plots'],
                        'xlabel': 'steps',
                        'ylabel': 'loss plots'
                    }
                )
            else:
                self.visualizer.line(
                    X=numpy.array([step_num]),
                    Y=numpy.array([v]),
                    win=self.loss_plots[self.keys['loss_plots']],
                    name=k, 
                    update='append',
                )

    def _plot_val_loss(self,
        metrics: typing.Dict[str, typing.Any],
        epoch: int,
        step_num: int
    ) -> None:
        for i, key in enumerate(metrics.keys()):
            value = metrics[key]
            if key not in self.metric_plots.keys():
                self.metric_plots[key] = self.visualizer.line(
                    X=numpy.array([step_num,step_num]),
                    Y=numpy.array([value,value]),
                    env=self.name,
                    name=key,
                    opts={
                        'linecolor': numpy.array([self.colors[2 + i]]), # second color and on
                        'legend': [k for k in metrics.keys() if k == key],
                        'title': key,
                        'xlabel':'steps',
                        'ylabel': key
                    }
                )
            else:
                self.visualizer.line(
                    X=numpy.array([step_num]),
                    Y=numpy.array([value]),
                    win=self.metric_plots[key],
                    name=key, 
                    update='append',
                )
            if key not in self.best_epoch_val_loss.keys():
                self.best_epoch_val_loss[key] = []
                self.best_epoch_val_step[key] = []
            while len(self.best_epoch_val_loss[key]) <= epoch:
                self.best_epoch_val_loss[key].append(numpy.Infinity)
                self.best_epoch_val_step[key].append(step_num)
            if value < self.best_epoch_val_loss[key][epoch]:
                self.best_epoch_val_loss[key][epoch] = value
                self.best_epoch_val_step[key][epoch] = step_num
                ind = numpy.isfinite(self.best_epoch_val_loss[key])
                self.visualizer.line(
                    X=numpy.array(self.best_epoch_val_step[key])[ind],
                    Y=numpy.array(self.best_epoch_val_loss[key])[ind],
                    update='insert',
                    win=self.metric_plots[key],
                    name='val_' + self.keys['epoch'],
                    opts={
                        'markers': True,
                        'linecolor': numpy.array([self.colors[2 + i]])
                    }
                )

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_metrics(self, 
        metrics:        typing.Dict[str, typing.Any],
        step:           int
    ) -> None:
        train_metrics = toolz.keymap(lambda k: k.replace('train_', ''), 
            toolz.keyfilter(lambda k: k.startswith('train_'), metrics)
        )
        val_metrics = toolz.keymap(lambda k: k.replace('val_', ''), 
            toolz.keyfilter(lambda k: k.startswith('val_'), metrics)
        )
        if train_metrics:
            epoch = int(metrics['epoch'])
            loss = float(metrics['total_loss'])
            self._plot_train_loss(epoch, loss, step)
            self._plot_losses(train_metrics, epoch, step)
        elif 'test' in metrics:
            return #TODO: test case 
        if val_metrics:
            epoch = int(metrics['epoch'])
            self._plot_val_loss(val_metrics, epoch, step)

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_hyperparams(self,
        params: typing.Dict[str, typing.Any] #TODO: or namespace object ?
    ) -> None:
        """Record hyperparameters
            :param params: argparse.Namespace containing the hyperparameters
        """
        self.visualizer.text(json2html.convert(json=dict(params)))

    @pytorch_lightning.loggers.base.rank_zero_only
    def save(self) -> None:
        """Save log data"""
        # self.visualizer.save(self.name)
        pass #TODO: support saving ?

    @pytorch_lightning.loggers.base.rank_zero_only
    def finalize(self, 
        status: str
    ) -> None:
        """Do any processing that is necessary to finalize an experiment
            :param status: Status that the experiment finished with (e.g. success, failed, aborted)
        """
        self.save()
        if self.clear_window:
            self.visualizer.close()

    @property
    def rank(self) -> int:
        """
            Process rank. In general, metrics should only be logged by the process
                with rank 0
        """
        return self._rank

    @rank.setter
    def rank(self, value: int) -> None:
        """Set the process rank"""
        self._rank = value

    @property
    def version(self) -> int:
        """Return the experiment version"""
        return self._version

    @property
    def experiment(self) -> typing.Any:
        return self.name


from moai.data.augmentation.albumentations.augmentations.random_gamma import RandomGamma
from moai.data.augmentation.albumentations.augmentations.random_sized_crop import RandomSizedCrop

__all__ = [
    "RandomGamma",
    "RandomSizedCrop",
]


from moai.visualization.visdom.base import Base
from moai.utils.arguments import ensure_string_list

import torch
import visdom
import functools
import typing
import logging
import numpy as np
import cv2
import colour

log = logging.getLogger(__name__)

__all__ = ["Gizmo2d"]

class Gizmo2d(Base):
    def __init__(self,
        images:         typing.Union[str, typing.Sequence[str]],
        gizmos:         typing.Union[str, typing.Sequence[str]],
        gt:             typing.Union[str, typing.Sequence[str]],
        pred:           typing.Union[str, typing.Sequence[str]],
        coords:         typing.Union[str, typing.Sequence[str]],
        color_gt:       typing.Union[str, typing.Sequence[str]],
        color_pred:     typing.Union[str, typing.Sequence[str]],
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,
        reverse_coords: bool=False,
    ):
        super(Gizmo2d, self).__init__(name, ip, port)
        self.images = ensure_string_list(images)
        self.gizmos = ensure_string_list(gizmos)
        self.gt = ensure_string_list(gt)
        self.pred = ensure_string_list(pred)
        self.color_gt = list(map(colour.web2rgb, ensure_string_list(color_gt)))
        self.color_pred = list(map(colour.web2rgb, ensure_string_list(color_pred)))
        self.coords = ensure_string_list(coords)
        self.reverse = reverse_coords
        self.gizmo_render = {
            'marker_circle': functools.partial(self.__draw_markers, 
                self.visualizer, marker=-1),
            'marker_diamond': functools.partial(self.__draw_markers, 
                self.visualizer, marker=cv2.MARKER_DIAMOND),
            'marker_star': functools.partial(self.__draw_markers, 
                self.visualizer, marker=cv2.MARKER_STAR),
            'marker_cross': functools.partial(self.__draw_markers, 
                self.visualizer, marker=cv2.MARKER_CROSS),
            'marker_square': functools.partial(self.__draw_markers, 
                self.visualizer, marker=cv2.MARKER_SQUARE),
            'bbox2d': functools.partial(self.__draw_2dbox, self.visualizer),
            'bbox3d': functools.partial(self.__draw_3dbox, self.visualizer),
            'axes': functools.partial(self.__draw_axes, self.visualizer),
            #TODO: axes to only receive a scale parameter and have the axis points hardcoded here
        }
        self.xforms = { #TODO: extract these into a common module
            'ndc': lambda coord, img: torch.addcmul(
                torch.scalar_tensor(0.5).to(coord), coord, torch.scalar_tensor(0.5).to(coord)
            ) * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
            'coord': lambda coord, img: coord,
            'norm': lambda coord, img: coord * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
        }

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for img, gzm, gt, pred, gt_c, pred_c, coord in zip(
            self.images, self.gizmos, self.gt, self.pred,
            self.color_gt, self.color_pred, self.coords
        ):
            gt_coord = tensors[gt].detach()
            pred_coord = tensors[pred].detach()
            if self.reverse:
                gt_coord = gt_coord.flip(-1)
                pred_coord = pred_coord.flip(-1)
            image = tensors[img].detach()
            self.gizmo_render[gzm](
                image,
                self.xforms[coord](gt_coord, image),
                self.xforms[coord](pred_coord, image),
                np.uint8(np.array(list(reversed(gt_c))) * 255),
                np.uint8(np.array(list(reversed(pred_c))) * 255),             
                coord, img, img, self.name
            )
    
    @staticmethod
    def __draw_2dbox(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_coordinates:     torch.Tensor,
        pred_coordinates:   torch.Tensor,
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str
    ) -> None:
        b, c, h, w = images.size()
        imgs = np.zeros(images.shape, dtype=np.uint8)
        gt_coords = gt_coordinates.detach().cpu()
        pred_coords = pred_coordinates.detach().cpu()
        gt_coords = gt_coords.numpy()
        pred_coords = pred_coords.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        line_size = int(0.005 * diagonal) #TODO: extract percentage param to config?

        for i in range(imgs.shape[0]):
            img = images[i, ...].detach().cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8)  
            for coords, color in zip(
                [gt_coords, pred_coords],
                [gt_color, pred_color]
            ):
                coord_i = coords[i, ...]
                pt1_x , pt1_y , w , h = coord_i
                pt2_x= pt1_x + w   # bottom right
                pt2_y = pt1_y + h   # bottom right
                cv2.rectangle(img,
                    (pt1_x, pt1_y),
                    (pt2_x, pt2_y),
                    color.tolist(),
                    line_size
                )
                imgs[i, ...] = img.transpose(2, 0, 1)
                del coord_i


        visdom.images(
            np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )
    
    @staticmethod
    def __draw_3dbox(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_coordinates:     torch.Tensor,
        pred_coordinates:   torch.Tensor,
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str
    ) -> None:
        b , c , h , w = images.size()
        imgs = np.zeros(images.shape, dtype=np.uint8)
        gt_coords = gt_coordinates.cpu()
        pred_coords = pred_coordinates.cpu()
        gt_coords = torch.flip(gt_coords, dims=[-1])
        pred_coords = torch.flip(pred_coords, dims=[-1])
        gt_coords = gt_coords.numpy()
        pred_coords = pred_coords.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        line_size = int(0.005 * diagonal) #TODO: extract percentage param to config?
        
        for i in range(imgs.shape[0]):
            img = images[i, ...].detach().cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8)  
            for coords, color in zip(
                [gt_coords, pred_coords],
                [gt_color, pred_color]
            ):
                coord_i = coords[i, ...]
                for k, key_ in enumerate(coord_i):
                    if (k + 1) % 2 == 0:
                        cv2.line(img, (int(point_1_x), int(point_1_y)), (int(coord_i[k][0]), int(coord_i[k][1])), color.tolist(), line_size)
                    else:
                        point_1_x = coord_i[k][0]
                        point_1_y = coord_i[k][1]
                for k, key_ in enumerate(coord_i):
                    if k == 0 or k == 1 or k ==4 or k == 5:
                        point_2_x = coord_i[k+2][0]
                        point_2_y = coord_i[k+2][1]
                        cv2.line(img, (int(point_2_x), int(point_2_y)), (int(coord_i[k][0]), int(coord_i[k][1])), color.tolist(), line_size)
                for k, key_ in enumerate(coord_i):
                    if k == 0 or k == 1 or k == 2 or k == 3:
                        point_2_x = coord_i[k+4][0]
                        point_2_y = coord_i[k+4][1]
                        cv2.line(img, (int(point_2_x), int(point_2_y)), (int(coord_i[k][0]), int(coord_i[k][1])), color.tolist(), line_size)
                imgs[i, ...] = img.transpose(2, 0, 1)
                del coord_i

        visdom.images(
            np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )

    @staticmethod
    def __draw_axes(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_axes:            torch.Tensor,
        pred_axes:          torch.Tensor,
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str
    ) -> None:
        b , c , h , w = images.size()
        imgs = np.zeros(images.shape, dtype=np.uint8)
        gt_axes = gt_axes.cpu()
        pred_axes = pred_axes.cpu()
        gt_axes = torch.flip(gt_axes, dims=[-1])
        pred_axes = torch.flip(pred_axes, dims=[-1]) 
        gt_axes = gt_axes.numpy()
        pred_axes = pred_axes.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        line_size = int(0.005 * diagonal)
        for i in range(imgs.shape[0]):
            img = images[i, ...].detach().cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8)
            for j , (coords, color) in enumerate(zip(
                [gt_axes, pred_axes],
                [gt_color, pred_color]
            )):
                coord_i = np.int32(coords[i, ...])
                #draw lines
                if j == 1:
                    alpha = 0.4
                    overlay = img.copy()
                    cv2.arrowedLine(overlay,tuple(coord_i[0]),tuple(coord_i[1]), (0,0,255), line_size)
                    cv2.arrowedLine(overlay,tuple(coord_i[0]),tuple(coord_i[2]), (255,0,0), line_size)
                    cv2.arrowedLine(overlay,tuple(coord_i[0]),tuple(coord_i[3]), (0,255,0), line_size)
                    cv2.addWeighted(overlay, alpha, img, 1 - alpha,
		                0, img)
                else:
                    cv2.arrowedLine(img,tuple(coord_i[0]),tuple(coord_i[1]), (0,0,255), line_size)
                    cv2.arrowedLine(img,tuple(coord_i[0]),tuple(coord_i[2]), (255,0,0), line_size)
                    cv2.arrowedLine(img,tuple(coord_i[0]),tuple(coord_i[3]), (0,255,0), line_size)
                
                imgs[i, ...] = img.transpose(2, 0, 1)
                del coord_i
        
        
        visdom.images(
            np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        ) 
    
    @staticmethod
    def __draw_markers(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_coordinates:     torch.Tensor,
        pred_coordinates:   torch.Tensor,
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str,
        marker:             int,
    ):
        imgs = np.zeros([images.shape[0], 3, images.shape[2], images.shape[3]], dtype=np.uint8)
        gt_coords = gt_coordinates.cpu()
        pred_coords = pred_coordinates.cpu()
        gt_coords = torch.flip(gt_coords, dims=[-1])
        pred_coords = torch.flip(pred_coords, dims=[-1])
        gt_coords = gt_coords.numpy()
        pred_coords = pred_coords.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        marker_size = int(0.02 * diagonal) #TODO: extract percentage param to config?
        line_size = int(0.005 * diagonal) #TODO: extract percentage param to config?
        for i in range(imgs.shape[0]):
            img = images[i, ...].cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8) if img.shape[2] > 1 else cv2.cvtColor(img.copy().astype(np.uint8), cv2.COLOR_GRAY2RGB)
            for coords, color in zip(
                [gt_coords, pred_coords],
                [gt_color, pred_color]
            ):       
                coord_i = np.int32(coords[i, ...])
                for k, coord in enumerate(coord_i):
                    if marker < 0:
                        cv2.circle(img, 
                            tuple(coord), 
                            marker_size, color.tolist(), thickness=line_size
                        )
                    else:
                        cv2.drawMarker(img, 
                            tuple(coord),
                            color.tolist(),
                            marker, marker_size, line_size
                        )
                imgs[i, ...] = img.transpose(2, 0, 1)
        visdom.images(
            np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )


import torch
import typing

__all__ = ['NoOp']

class Null(torch.optim.Optimizer):
  def __init__(self,
    parameters: typing.Iterator[torch.nn.Parameter],
  ):
    super(Null, self).__init__(parameters, {"lr": 0.0, "eps": 1e-8})

  def step(self, closure=None):
    if closure is not None:
      closure()
    return None

class NoOp(object): 
  def __init__(self, 
      parameters: typing.Iterator[torch.nn.Parameter],
  ):
    self.optimizers = [Null(parameters)]

  def step(self, closure=None):
    return None


import moai.nn.convolution as mic
import moai.nn.activation as mia

import torch

__all__ = [
    "Standard",
    "PreResidual",
    "PreActivation",
]

'''
    Slightly adapted version of 
    Deep Residual Learning for Image Recognition (https://arxiv.org/pdf/1512.03385.pdf)
    (adaptation on activation ordering as denoted in the factory below)
'''
class Standard(torch.nn.Module): # (b) in https://towardsdatascience.com/an-overview-of-resnet-and-its-variants-5281e2f56035
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(Standard, self).__init__()
        self.W1 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=in_features,
            out_channels=out_features,
            stride=2 if strided else 1,
            **convolution_params
        )
        self.A1 = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W2 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=out_features,
            out_channels=out_features,
            **convolution_params
        )
        self.A2 = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )
        self.S = torch.nn.Identity() if in_features == out_features\
            else mic.make_conv_1x1(
                    convolution_type=convolution_type,
                    in_channels=in_features,
                    out_channels=out_features
                # using a 3x3 conv for shortcut downscaling instead of a 1x1 (used in detectron2 for example)
                ) if not strided else mic.make_conv_3x3(
                    convolution_type=convolution_type,
                    in_channels=in_features,
                    out_channels=out_features,
                    stride=2
                )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.W2(self.A1(self.W1(x)))  # y = W2 * A1(W1 * x)
        return self.A2(self.S(x) + y)     # out = A2(S(x) + y)

'''
    Slightly adapted version of
    Identity Mappings in Deep Residual Networks (https://arxiv.org/pdf/1603.05027.pdf)
    (adaptation on activation ordering as denoted in the factory below)
'''
class PreResidual(Standard): # (c) in https://towardsdatascience.com/an-overview-of-resnet-and-its-variants-5281e2f56035
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(PreResidual, self).__init__(
            convolution_type=convolution_type,
            activation_type=activation_type,
            in_features=in_features,
            out_features=out_features,
            convolution_params=convolution_params,
            activation_params=activation_params,
            strided=strided,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.A2(self.W2(self.A1(self.W1(x))))       # y = A2(W2 * A1(W1 * x))
        return self.S(x) + y                            # out = x + y

'''
    Slightly adapted version of
    Identity Mappings in Deep Residual Networks (https://arxiv.org/pdf/1603.05027.pdf)
    (adaptation on activation ordering as denoted in the factory below)
'''
class PreActivation(torch.nn.Module): # (e) in https://towardsdatascience.com/an-overview-of-resnet-and-its-variants-5281e2f56035
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(PreActivation, self).__init__()
        self.W1 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=in_features,
            out_channels=out_features,
            stride=2 if strided else 1,
            **convolution_params
        )
        self.A1 = mia.make_activation(
            features=in_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W2 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=out_features,
            out_channels=out_features,
            **convolution_params
        )
        self.A2 = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )
        self.S = torch.nn.Identity() if in_features == out_features\
            else mic.make_conv_1x1(
                convolution_type=convolution_type,
                in_channels=in_features,
                out_channels=out_features
                # using a 3x3 conv for shortcut downscaling instead of a 1x1 (used in detectron2 for example)
            ) if not strided else mic.make_conv_3x3(
                convolution_type=convolution_type,
                in_channels=in_features,
                out_channels=out_features,
                stride=2
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.W2(self.A2(self.W1(self.A1(x))))       # y = W2 * A2(W1 * A1(x))
        return self.S(x) + y                            # out = x + y


import torch
import typing

#NOTE: from https://github.com/jxbz/madam 

class SignSGD(torch.optim.Optimizer):
    r"""Implements the Sign SGD optimisation algorithm.

        ??? info "[signSGD: Compressed Optimisation for Non-Convex Problems](https://arxiv.org/pdf/1802.04434.pdf)"
            **Abstract**: Training large neural networks requires distributing learning across multiple workers, where the
            cost of communicating gradients can be a significant bottleneck. SIGNSGD alleviates this problem by transmitting just the sign of each minibatch
            stochastic gradient. We prove that it can get the best of both worlds: compressed gradients and
            SGD-level convergence rate. The relative `1/`2 geometry of gradients, noise and curvature informs whether SIGNSGD or SGD is theoretically
            better suited to a particular problem. On the practical side we find that the momentum counterpart
            of SIGNSGD is able to match the accuracy and convergence speed of ADAM on deep Imagenet
            models. We extend our theory to the distributed setting, where the parameter server uses majority
            vote to aggregate gradient signs from each worker enabling 1-bit compression of worker-server communication in both directions. Using a theorem
            by Gauss (1823) we prove that majority vote can achieve the same reduction in variance as full
            precision distributed SGD. Thus, there is great promise for sign-based optimisation schemes to
            achieve fast communication and fast convergence. Code to reproduce experiments is to be found at
            https://github.com/jxbz/signSGD.

        ??? info "[GitHub @ jxbz/madam](https://github.com/jxbz/madam)"

        Arguments:
            lr (float, required): The learning rate.
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float
    ):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        defaults = dict(lr=lr)
        super(SignSGD, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                sign = torch.sign(p.grad.data)
                p.data.add_(-group['lr'], sign)

        return loss


import torch

import typing
import functools

__all__ = [
    "NormToCoords",
    "NormToNdc",
    "NdcToCameraCoords",
    "ScaleCoords",
    "UpscaleCoords_x2",
    "UpscaleCoords_x4",
    "DownscaleCoords_x2",
    "DownscaleCoords_x4",
    "CoordsToNorm",
]

#TODO: extract to generation/grid/conversions and refactor to support all cases with from/to arguments
class NormToCoords(torch.nn.Module):
    def __init__(self,
        mode: str="coord",
        flip: bool= False,
    ):
        super(NormToCoords,self).__init__()
        self.mode = mode
        self.flip = flip

    def forward(self, coords: torch.Tensor, img: torch.Tensor) -> torch.Tensor:
        if self.flip:
            coords = coords * torch.Tensor([*img.shape[2:]]).flip(-1).to(coords).expand_as(coords)
        else:
            coords = coords * torch.Tensor([*img.shape[2:]]).to(coords).expand_as(coords)
        return coords

class NormToNdc(torch.nn.Module):
    def __init__(self,
        mode: str="coord",
        flip: bool= False,
    ):
        super(NormToNdc,self).__init__()
        self.mode = mode
        self.flip = flip

    def forward(self,coords: torch.Tensor, img: torch.Tensor) -> torch.Tensor:
        if self.flip:
            ndc = torch.addcmul(torch.scalar_tensor(-1.0).to(coords), coords, torch.scalar_tensor(2.0).to(coords)).flip(-1), # from [0, 1] to [-1, 1]
        else:
            ndc = torch.addcmul(torch.scalar_tensor(-1.0).to(coords), coords, torch.scalar_tensor(2.0).to(coords)), # from [0, 1] to [-1, 1]
        return ndc[0]


class NdcToCameraCoords(torch.nn.Module):
    def __init__(
        self,
        fov:    float=64.69, # in degrees
        width:  int=320,
        height: int=240,
        flip:   bool= False,
        order:  str='xy' , # order of coords
    ):
        super(NdcToCameraCoords,self).__init__()
        self.flip = flip
        self.order = order
        self.tanfov = torch.tan((torch.deg2rad(torch.tensor(fov) / 2.0).float()))
        self.aspect_ratio = width / height
    
    def forward(self,coords: torch.Tensor) -> torch.Tensor:
        if self.order == "xy":
            camera_coords = torch.div(coords,torch.tensor([self.aspect_ratio / self.tanfov,(1 / self.tanfov)]).to(coords))
        else:
            camera_coords = torch.div(coords,torch.tensor([(1 / self.tanfov),self.aspect_ratio / self.tanfov]).to(coords))
        return camera_coords if not self.flip else camera_coords.flip(-1)

class ScaleCoords(torch.nn.Module):
    def __init__(self,
        scale:  float=0.5,
        flip:   bool= False,
    ):
        super(ScaleCoords,self).__init__()
        self.scale = scale
        self.flip = flip

    def forward(self,
        coords: torch.Tensor 
    ) -> torch.Tensor:
        coords = self.scale * coords
        if self.flip:
            return coords.flip(-1)
        else:
            return coords


class CoordsToNorm(torch.nn.Module):
    def __init__(self,
        flip: bool= False,
    ):
        super(CoordsToNorm,self).__init__()
        self.flip = flip

    def forward(self,
        coords: torch.Tensor,
        grid:   typing.Optional[torch.Tensor]=None,
        image:  typing.Optional[torch.Tensor]=None,
    ) -> torch.Tensor:
        dims_from = grid if grid is not None else image
        dims = torch.tensor(dims_from.shape[2:], dtype=torch.float32, device=coords.device)
        if self.flip:
            dims = dims.flip(-1)
        return coords / dims.expand_as(coords)

UpscaleCoords_x2 = functools.partial(
    ScaleCoords, 
    scale=2
)

UpscaleCoords_x4 = functools.partial(
    ScaleCoords, 
    scale=4
)

DownscaleCoords_x2 = functools.partial(
    ScaleCoords, 
    scale=0.5
)

DownscaleCoords_x4 = functools.partial(
    ScaleCoords, 
    scale=0.25
)


import torch.nn as ptnn
import moai.nn.utils as miu

import functools
import typing
import logging

log = logging.getLogger(__name__)

__all__ = [
    "make_linear_op",
    "make_linear_block",
    "make_linear_relu_bn",
    "make_linear_elu",
    "make_linear_sigmoid",
    "make_linear_tanh",
    "make_linear_no_activation",
]

__LINEAR_FACTORY__ = {
    "linear":               ptnn.Linear,
    "bilinear":             ptnn.Bilinear,
}

def _update_linear_op(name: str, type: typing.Type):
    if name not in __LINEAR_FACTORY__.keys():
        __LINEAR_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the linear operation factory.")

def make_linear_op(
    linear_type: str,
    in_features: int,
    out_features: int,
    bias: bool=True,
    **kwargs
) -> ptnn.Module:
    if linear_type in __LINEAR_FACTORY__.keys():
        other_kwargs = {}
        if linear_type == "bilinear" and "in1_features" not in kwargs:
            other_kwargs.update({"in1_features": in_features})
        return miu.instantiate(__LINEAR_FACTORY__[linear_type], 
        **{
            **locals(),
            **kwargs,
            **other_kwargs,
        })
    else:
        log.error(f"Linear type ({linear_type}) not found.")

__LINEAR_BLOCK_FACTORY__ = {
    
}

def _update_linear_block(name: str, type: typing.Type):
    if name not in __LINEAR_BLOCK_FACTORY__.keys():
        __LINEAR_BLOCK_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the linear block factory.")

def make_linear_block(
        block_type: str,
        linear_type: str,#TODO: obsolete, to remove
        in_features: int,
        out_features: int,
        activation_type: str,
        linear_params: dict={},
        activation_params: dict={"inplace": True},
        **kwargs
    ) -> ptnn.Module:
        if block_type in __LINEAR_BLOCK_FACTORY__.keys():
            return miu.instantiate(__LINEAR_BLOCK_FACTORY__[block_type],
                **{ 
                    **locals(),
                    "linear_params": linear_params, #TODO: merge /w kwargs?
                    "activation_params": activation_params,
                    **kwargs #TODO: merge to linear_params if otherwise unused?                    
                })
        else:
            log.error(f"Linear block type ({block_type}) not found.")

make_linear_relu_bn = functools.partial(make_linear_block,
    block_type="linear",
    linear_type="linear",
    activation_type="relu_bn",
)

make_linear_elu = functools.partial(make_linear_block,
    block_type="linear",
    linear_type="linear",
    activation_type="elu",
)

make_linear_sigmoid = functools.partial(make_linear_block,
    block_type="linear",
    linear_type="linear",
    activation_type="sigmoid",
)

make_linear_tanh = functools.partial(make_linear_block,
    block_type="linear",
    linear_type="linear",
    activation_type="tanh",
)

import moai.nn.linear.torch as ptln

if "linear" not in __LINEAR_BLOCK_FACTORY__:
    _update_linear_block("linear", ptln.LinearBlock)    


import functools
import typing
import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["Mask"]

class Mask(torch.nn.Module):
    def __init__(self,
        value:      float=0.0,
        inverse:    bool=True,
    ):
        super(Mask, self).__init__()
        self.value = value
        self.inverse = inverse

    def forward(self,
        tensor: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        m = ~mask if self.inverse else mask
        m_c = mask.shape[1]
        t_c = tensor.shape[1]
        if m_c > t_c:
            for i in range(m_c):
                tensor[m[:, i, ...]] = self.value
            tensor[~mask] = self.value
        elif t_c > m_c:
            for i in range(t_c):
                tensor[:, i, ...].unsqueeze(1)[m] = self.value
        else:
            tensor[m] = self.value
        return tensor

class Index(torch.nn.Module):
    def __init__(self,
        indices:    typing.Sequence[int],
        dim:        int=1,        
    ):
        super(Index, self).__init__()
        self.index = functools.partial(torch.index_select, dim=dim)
        self.register_buffer("indices", torch.tensor(list(indices), dtype=torch.long))

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
            return self.index(input=tensor, index=self.indices)



from moai.parameters.initialization import Cascade
from moai.utils.torch import get_submodule

import torch
import omegaconf.omegaconf
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ["Named"]

class Named(Cascade):
    def __init__(self,         
        schemes: omegaconf.DictConfig={},
        modules: omegaconf.DictConfig={},        
    ):
        super(Named, self).__init__(schemes)
        self.inits = modules

    def __call__(self,
        module: torch.nn.Module
    ) -> None:
        found = []
        for k, v in self.inits.items():
            m = get_submodule(module, k)
            if m is not None:
                found += [k]
                log.info(f"Initializing {k} from {v.ckpt} " + ("strictly." if v.strict else "relaxed."))
                data = torch.load(v.ckpt, map_location=lambda s, l: s)
                if v.source:
                    split = v.source.split('.')
                    data = toolz.get_in(split, data) or data
                if v.replace:
                    for src, tgt in v.replace.items():
                        data = toolz.keymap(lambda k: k.replace(src, tgt), data)
                m.load_state_dict(data, strict=v.strict or False)
        for k in found:
            self.inits.pop(k)
        super(Named, self).__call__(module)



from moai.monads.utils import expand_spatial_dims
from moai.monads.distribution import CenterOfMass

import torch

__all__ = ["VarianceRegularization"]

__KEYPOINTS_TO_COORDS__ = {#TODO: create grid conversion OPs and link to there
    'ndc': lambda coord, img: torch.addcmul(
        torch.scalar_tensor(0.5).to(coord), coord, torch.scalar_tensor(0.5).to(coord)
    ) * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
    'coord': lambda coord, img: coord,
    'norm': lambda coord, img: coord * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
}

__GRID_TO_COORDS__ = {#TODO: create grid conversion OPs and link to there
    'ndc': lambda grid: torch.addcmul(
        torch.scalar_tensor(0.5).to(grid), grid, torch.scalar_tensor(0.5).to(grid)
    ) * expand_spatial_dims(torch.Tensor([*grid.shape[2:]]).to(grid), grid),
    'coord': lambda grid: grid,
    'norm': lambda grid: grid * expand_spatial_dims(torch.Tensor([*grid.shape[2:]]).to(grid), grid),
}

class VarianceRegularization(CenterOfMass):
    def __init__(self,
        sigma: float=1.5, # in pixels
        grid_type: str='ndc', # 'ndc', 'coord', 'norm'        
    ):
        super(VarianceRegularization, self).__init__(mode='coords', flip=False)
        self.target_variance = sigma ** 2
        self.grid_type = grid_type

    def forward(self, 
        heatmaps:   torch.Tensor,        # [B, K, (D), (H), W]
        grid:       torch.Tensor,        # [B, (1-3), (D), (H), W]
        keypoints:  torch.Tensor,        # [B, K, (1-3)]
    ) -> torch.Tensor:        
        k = __KEYPOINTS_TO_COORDS__[self.grid_type](keypoints)        
        grid = __GRID_TO_COORDS__[self.grid_type](grid)
        diffs = (grid.unsqueeze(1) - k) ** 2 # [B, K, (1-3), (D), (H), W]
        pred_stds = super(VarianceRegularization, self).forward(diffs, heatmaps)
        pred_variances = pred_stds ** 2
        squared_error = (pred_variances - self.target_variance) ** 2
        return torch.sum(squared_error, dim=-1)





import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.utils as miu
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.sampling.spatial.upsample as mius

import torch
import typing
import omegaconf.omegaconf

__all__ = ["StartTransition", "StageTransition"]

class StartTransition(torch.nn.Module):
    def __init__(self, 
        in_features:           int,
        start_features:        int,
        identity:               omegaconf.DictConfig,
        branched:               omegaconf.DictConfig,
    ):
        super(StartTransition, self).__init__()        
        self.conv_id = mic.make_conv_block(
            block_type="conv2d",
            convolution_type=identity.convolution,
            in_features=in_features,
            out_features=start_features,
            activation_type=identity.activation,
            convolution_params={
                'kernel_size':  identity.kernel_size,
                'stride':       identity.stride,
                'padding':      identity.padding
            }
        )
        self.downscale_br = mids.make_downsample(
            downscale_type=branched.downscale,
            features=in_features,
            kernel_size=3 if branched.downscale == 'maxpool2d_aa' else 2,
        )
        self.conv_br = mic.make_conv_block(
            block_type="conv2d",
            convolution_type=branched.convolution,
            in_features=in_features,
            out_features=start_features * 2,
            activation_type=branched.activation,
            convolution_params={
                'kernel_size':  branched.kernel_size,
                'stride':       branched.stride,
                'padding':      branched.padding
            }
        )

    def forward(self, x: torch.Tensor) -> typing.List[torch.Tensor]:
        return [self.conv_id(x), self.conv_br(self.downscale_br(x))]

class StageTransition(torch.nn.Module):
    def __init__(self, 
        branches:               int,
        prev_branch_features:   int,        
        branched:               omegaconf.DictConfig,
    ):
        super(StageTransition, self).__init__()
        self.list = torch.nn.ModuleList([
            *(torch.nn.Identity() for _ in range(branches)), 
            torch.nn.Sequential(
                mids.make_downsample(
                    downscale_type=branched.downscale,
                    features=prev_branch_features,
                    kernel_size=3 if branched.downscale == 'maxpool2d_aa' else 2,
                ),
                mic.make_conv_block(
                    block_type="conv2d",
                    convolution_type=branched.convolution,
                    in_features=prev_branch_features,
                    out_features=prev_branch_features * 2,
                    activation_type=branched.activation,
                    convolution_params={
                        'kernel_size':  branched.kernel_size,
                        'stride':       branched.stride,
                        'padding':      branched.padding
                    },
                )
            )
        ])

    def forward(self, 
        branches: typing.Union[typing.Tuple[torch.Tensor, ...], typing.List[torch.Tensor]]
    ) -> typing.List[torch.Tensor]:
        return [m(t) for t, m in zip(branches + [branches[-1]], self.list)]


import math
import torch
import torch
import typing

__all__ = ['AdamAIO']

#NOTE: modified from https://github.com/kayuksel/pytorch-adamaio

class AdamAIO(torch.optim.Optimizer):
    """Implements the All-In-One Adam Optimizer.

        The All-In-One Adam Optimizer, where several novelties are combined from following papers:

        - [Decoupled Weight Decay Regularization for Adam](https://arxiv.org/abs/1711.05101)

            Authors shown that the real reason why Momentum optimizer is often outperforming Adam in generalization was due to 
            the fact that Adam does not perform well under L2 regularization and developed decoupled weight decay as a solution.

        - [Online Learning Rate Adaptation with Hypergradient Descent](https://arxiv.org/abs/1703.04782)

            This is enabled via "hypergrad" parameter by setting it to any value except zero. It enables the optimizer to update
            the learning-rate itself by the technique proposed in the paper, instead of giving an external schedule which would 
            require lots of additional hyperparameters. It is especially useful when one doesn't want to hypertune a schedule.

        - [Closing the Generalization Gap of Adaptive Gradient Methods in Training Deep Neural Networks](https://arxiv.org/abs/1711.05101)

            This can be set by the "partial" parameter, which controls how likely the optimizer acts similar to Adam (1.0) and 
            SGD (0.0), which is very useful if hypertuned. One can also update (decay) this parameter online to switch between 
            Adam and SGD optimizers in an easy way, which has been recommended by previous research for a better generalization.
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=2e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=5e-3,
        weight_decay: float=5e-6,
        hypergrad: float=1e-7,
        partial: float=0.75,
    ):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, hypergrad=hypergrad, partial=partial)
        super().__init__(params, defaults)
    def step(self, closure=None):
        loss = None if closure is None else closure()
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None: continue
                grad = p.grad.data
                state = self.state[p]
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                state['step'] += 1
                if group['hypergrad'] > 0 and state['step'] > 1:
                    prev_bias_correction1 = 1 - beta1 ** (state['step'] - 1)
                    prev_bias_correction2 = 1 - beta2 ** (state['step'] - 1)
                    h = torch.dot(grad.view(-1), torch.div(exp_avg, exp_avg_sq.sqrt().add_(group['eps'])).view(-1)) * math.sqrt(prev_bias_correction2) / prev_bias_correction1
                    group['lr'] += group['hypergrad'] * h
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                denom = exp_avg_sq.sqrt().add_(group['eps'])
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1
                if group['weight_decay'] != 0:
                    decayed_weights = torch.mul(p.data, group['weight_decay'])
                    p.data.addcdiv_(-step_size, exp_avg, denom**group['partial'])
                    p.data.sub_(decayed_weights)
                else:
                    p.data.addcdiv_(-step_size, exp_avg, denom**group['partial'])
        return loss



import torch
import typing

__all__ = ['LARS']

class LARS(torch.optim.Optimizer):
    r"""Implements layer-wise adaptive rate scaling for SGD.

    - **Paper**: [Large Batch Training of Convolutional Networks](https://arxiv.org/abs/1708.03888)
    - **Implementation**: [GitHub @ noahgolmant](https://github.com/noahgolmant/pytorch-lars)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): base learning rate (\gamma_0)
        momentum (float, optional): momentum factor (default: 0) ("m")
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
            ("\beta")
        eta (float, optional): LARS coefficient
        max_epoch (int): maximum training epoch to determine polynomial LR decay.

    Example:
        >>> optimizer = LARS(model.parameters(), lr=0.1, eta=1e-3)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        #lr=torch.optim.optimizer.required,
        lr=None,
        momentum: float=0.9,
        weight_decay: float=.0005,
        eta: float=0.001,
        max_epoch: int=200
    ):
        #if lr is not torch.optim.optimizer.required and lr < 0.0:
        if lr is not None and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}"
                             .format(weight_decay))
        if eta < 0.0:
            raise ValueError("Invalid LARS coefficient value: {}".format(eta))

        self.epoch = 0
        defaults = dict(lr=lr, momentum=momentum,
                        weight_decay=weight_decay,
                        eta=eta, max_epoch=max_epoch)
        super(LARS, self).__init__(params, defaults)

    def step(self, epoch=None, closure=None):
        """Performs a single optimization step2.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
            epoch: current epoch to calculate polynomial LR decay schedule.
                if None, uses self.epoch and increments it.
        """
        loss = None
        if closure is not None:
            loss = closure()

        if epoch is None:
            epoch = self.epoch
            self.epoch += 1

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            eta = group['eta']
            lr = group['lr']
            max_epoch = group['max_epoch']

            for p in group['params']:
                if p.grad is None:
                    continue

                param_state = self.state[p]
                d_p = p.grad.data

                weight_norm = torch.norm(p.data)
                grad_norm = torch.norm(d_p)

                # Global LR computed on polynomial decay schedule
                decay = (1 - float(epoch) / max_epoch) ** 2
                global_lr = lr * decay

                # Compute local learning rate for this layer
                local_lr = eta * weight_norm / \
                    (grad_norm + weight_decay * weight_norm)

                # Update the momentum term
                actual_lr = local_lr * global_lr

                if 'momentum_buffer' not in param_state:
                    buf = param_state['momentum_buffer'] = \
                            torch.zeros_like(p.data)
                else:
                    buf = param_state['momentum_buffer']
                buf.mul_(momentum).add_(actual_lr, d_p + weight_decay * p.data)
                p.data.add_(-buf)

        return loss


import os
import logging

def assert_path(
    logger:             logging.Logger,
    name:               str,
    path:               str,
) -> None:
    if not os.path.exists(path):
        logger.error(f"Path ({path}) does not exist.")

def ensure_path(
    logger:             logging.Logger,
    name:               str,
    path:               str,
) -> None:
    if not path or not os.path.exists(path):
        logger.error(f"Path ({path}) does not exist, reverting to cwd ({os.getcwd()})")
        return os.getcwd()
    else:
        return path


from moai.utils.arguments import (
    ensure_path,
    ensure_choices,
)
from moai.monads.execution.cascade import _create_accessor

import trimesh
import torch
import typing
import logging
import os

__all__ = ["Mesh"]

log = logging.getLogger(__name__)

class Mesh(typing.Callable[[typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]], None]):

    __MODES__ = ['all', 'overwrite']
    __FORMATS__ = ['ply', 'obj']

    def __init__(self,
        path:              str,
        vertices:          typing.Union[str, typing.Sequence[str]],
        faces:             typing.Union[str, typing.Sequence[str]],
        filetype:          typing.Union[str, typing.Sequence[str]],
        mode:              str="overwrite", # all"
        format:            str="05d",
    ):
        self.mode = ensure_choices(log, "saving mode", mode, Mesh.__MODES__)
        self.folder = ensure_path(log, "output folder", path)
        self.formats = [ensure_choices(log, "output format", ext, Mesh.__FORMATS__) for ext in filetype]
        self.vertices = [vertices] if isinstance(vertices, str) else list(vertices)
        self.vertices = [_create_accessor(k) for k in self.vertices]
        self.faces = [faces] if isinstance(faces, str) else list(faces)
        self.faces = [_create_accessor(k) for k in self.faces]
        self.format = format
        log.info(f"Exporting meshes locally @ {self.folder}")
        self.index = 0

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for v, f, ext in zip(
            self.vertices, self.faces, self.formats
        ):
            verts = v(tensors).detach().cpu().numpy()
            indices = f(tensors).detach().cpu().numpy()
            bs = verts.shape[0]
            for i in range(bs):
                trimesh.Trimesh(
                    verts[i], indices[i], process=False
                ).export(os.path.join(
                    self.folder, f"mesh_{(self.index + i):{self.format}}.{ext}")
                )
        self.index = 0 if self.mode == "overwrite" else self.index + bs




from moai.parameters.optimization.optimizers.a2grad import (
    A2GradUni,
    A2GradInc,
    A2GradExp    
)
from moai.parameters.optimization.optimizers.acc_sgd import AccSGD
from moai.parameters.optimization.optimizers.adabelief import AdaBelief
from moai.parameters.optimization.optimizers.adabound import (
    AdaBound,
    AdaBoundW
)
from moai.parameters.optimization.optimizers.adafactor import AdaFactor
from moai.parameters.optimization.optimizers.adai import Adai
from moai.parameters.optimization.optimizers.adam_aio import AdamAIO
from moai.parameters.optimization.optimizers.adamod import AdaMod
from moai.parameters.optimization.optimizers.adamp import AdamP
from moai.parameters.optimization.optimizers.adamt import AdamT
from moai.parameters.optimization.optimizers.adamw import AdamW
from moai.parameters.optimization.optimizers.adashift import AdaShift
from moai.parameters.optimization.optimizers.adam_srt import (
    AdamSRT,
    AdamS
)
from moai.parameters.optimization.optimizers.adax import (
    AdaX,
    AdaXW
)
from moai.parameters.optimization.optimizers.aggmo import AggMo
from moai.parameters.optimization.optimizers.apollo import (
    Apollo,
    ApolloW
)
from moai.parameters.optimization.optimizers.centralization import (
    CentralizedAdam,
    CentralizedAdamW,
    CentralizedPlainRAdam,
    CentralizedRAdam,
    CentralizedRanger,
    CentralizedSGD
)
from moai.parameters.optimization.optimizers.diffgrad import DiffGrad
from moai.parameters.optimization.optimizers.fromage import Fromage
from moai.parameters.optimization.optimizers.hypergrad_sgd import HypergradSGD
from moai.parameters.optimization.optimizers.lamb import Lamb
from moai.parameters.optimization.optimizers.lbfgsls import LBFGS
from moai.parameters.optimization.optimizers.larc import LARCAdam
from moai.parameters.optimization.optimizers.lars import LARS
from moai.parameters.optimization.optimizers.madam import (
    Madam,
    IntegerMadam,
)
from moai.parameters.optimization.optimizers.novograd import (
    NovoGrad,
    # AdamW,
)
from moai.parameters.optimization.optimizers.pid import PID
from moai.parameters.optimization.optimizers.qhadam import QHAdam
from moai.parameters.optimization.optimizers.qhadamw import QHAdamW
from moai.parameters.optimization.optimizers.qhm import QHM
from moai.parameters.optimization.optimizers.radam import (
    RAdam,
    PlainRAdam,
    # AdamW
)
from moai.parameters.optimization.optimizers.ralamb import Ralamb
from moai.parameters.optimization.optimizers.ranger import Ranger
from moai.parameters.optimization.optimizers.rangerlars import RangerLars
from moai.parameters.optimization.optimizers.sa import (
    UniformSimulatedAnnealing,
    GaussianSimulatedAnnealing,
)
from moai.parameters.optimization.optimizers.sadam import SAdam
from moai.parameters.optimization.optimizers.sgdp import SGDP
from moai.parameters.optimization.optimizers.sgdw import SGDW
from moai.parameters.optimization.optimizers.shampoo import Shampoo
from moai.parameters.optimization.optimizers.sign_sgd import SignSGD
from moai.parameters.optimization.optimizers.swa import SWAdam
from moai.parameters.optimization.optimizers.swats import SWATS
from moai.parameters.optimization.optimizers.yogi import Yogi

__all__ = [
    "A2GradUni",
    "A2GradInc",
    "A2GradExp", 
    "AccSGD",
    "AdaBelief",
    "AdaBound",
    "AdaBoundW",
    "AdaFactor",
    "Adai",
    "AdamAIO",
    "AdaMod",
    "AdamP",
    "AdamT",
    "AdamW",
    "AdaShift",
    "AdamSRT",
    "AdamS",
    "AdaX",
    "AdaXW",
    "AggMo",
    "Apollo",
    "ApolloW",
    "CentralizedAdam",
    "CentralizedAdamW",
    "CentralizedPlainRAdam",
    "CentralizedRAdam",
    "CentralizedRanger",
    "CentralizedSGD",
    "DiffGrad",
    "HypergradSGD",
    "Fromage",
    "IntegerMadam",
    "Lamb",    
    "LARCAdam",
    "LARS",
    "LBFGS",
    "Lookahead", #TODO: create factory optimizer
    "Madam",
    "NovoGrad",
    "PlainRAdam",
    "PID",
    "QHAdam",
    "QHAdamW",
    "QHM",
    "RAdam",
    "Ralamb",
    "Ranger",
    "RangerLars",
    "UniformSimulatedAnnealing",
    "GaussianSimulatedAnnealing",
    "SAdam",
    "SGDP",
    "SGDW",
    "Shampoo",
    "SignSGD",
    "SWAdam", #TODO: create factory optimizer
    "SWATS",
    "Yogi",
]


from moai.export.local.image2d import Image2d
from moai.export.local.blend2d import Blend2d
from moai.export.local.pkl import Pkl
from moai.export.local.npz import Npz

__all__ = [
    "Blend2d",
    "Image2d",
    "Pkl",
    "Npz",
]


from collections.abc import Callable, Iterator
from moai.monads.execution.cascade import _create_accessor

import numpy as np
import npstreams
import typing
import torch
import logging

log = logging.getLogger(__name__)

__all__ = ['MeanStd']

class MeanStd(Callable):
    class Iter(Iterator):
        def __init__(self, proxy: typing.Any):
            self.proxy = proxy

        def __iter__(self):
            return self

        def __next__(self) -> typing.Dict[str, torch.Tensor]:
            return self.proxy.tensors
        
    def __init__(self,
        key:        typing.Union[str, typing.Sequence[str]],
        ddof:       int=0,
        ignore_nan: bool=False,
    ) -> None:
        super().__init__()
        self.names = [key] if isinstance(key, str) else list(key)
        self.ddof = ddof
        self.ignore_nan = ignore_nan
        self.keys = [_create_accessor(k) for k in self.names]    
        self.avg = []
        self.std = []
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        if not hasattr(self, 'tensors'):
            self.tensors = tensors
            for name, k, a in zip(self.names, self.keys):
                self.avg.append(npstreams.imean(map(lambda d: d.numpy(), 
                    npstreams.iload(MeanStd.Iter(self), k))),
                    ddof=self.ddof, ignore_nan=self.ignore_nan,
                )
                self.std.append(npstreams.istd(map(lambda d: d.numpy(), 
                        npstreams.iload(MeanStd.Iter(self), k))
                    ), ddof=self.ddof, ignore_nan=self.ignore_nan,
                )
            tensors[f"{name}_mean"] = torch.zeros_like(k(tensors))
            tensors[f"{name}_std"] = tensors[f"{name}_mean"]
            tensors[f"{name}_p1std"] = tensors[f"{name}_mean"]
            tensors[f"{name}_m1std"] = tensors[f"{name}_mean"]
        else:
            self.tensors = tensors
            for name, avg, std in zip(self.names, self.avg, self.std):
                for a, s in zip(next(avg), next(std)):
                    continue
                tensors[f"{name}_mean"] = torch.from_numpy(a)[np.newaxis, ...]
                tensors[f"{name}_std"] = torch.from_numpy(s)[np.newaxis, ...]
                tensors[f"{name}_p1std"] = torch.from_numpy(a + s)[np.newaxis, ...]
                tensors[f"{name}_m1std"] = torch.from_numpy(a - s)[np.newaxis, ...]


from moai.modules.lightning.highres.highres import HighResolution
from moai.modules.lightning.highres.transition import (
    StartTransition,
    StageTransition
)
from moai.modules.lightning.highres.head import (
    TopBranch as TopBranchHead,
    AllBranches as AllBranchesHead,
    Higher as HigherHead
)

__all__ = [
    "HighResolution",
    "StartTransition",
    "StageTransition",
    "TopBranchHead",
    "AllBranchesHead",
    "HigherHead",
]


import torch
import typing

__all__ = ['QHM']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/qhm.py

class QHM(torch.optim.Optimizer):

    GRAD = 'grad'
    DIRECT = 'direct'

    r"""Implements quasi-hyperbolic momentum (QHM)  optimization algorithm.

   - **Paper**: [QUASI-HYPERBOLIC MOMENTUM AND ADAM FOR DEEP LEARNING](https://arxiv.org/pdf/1810.06801.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        momentum: momentum factor (:math:`\beta` from the paper)
        nu: immediate discount factor (:math:`\nu` from the paper)
        weight_decay: weight decay (L2 regularization coefficient, times two)
            (default: 0.0)
        weight_decay_type: method of applying the weight decay:
            ``"grad"`` for accumulation in the gradient
            (same as :class:`torch.optim.SGD`) or
            ``"direct"`` for direct application to the parameters
            (default: ``"grad"``)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.QHM(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()


    __ https://arxiv.org/abs/1810.06801

    Note:
        Reference code: https://github.com/facebookresearch/qhoptim
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        momentum: float=0.0,
        nu: float=0.7,
        weight_decay: float=0.0,
        weight_decay_type: str='grad',
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if momentum < 0.0:
            raise ValueError('Invalid momentum value: {}'.format(momentum))
        if weight_decay < 0.0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )
        if weight_decay_type not in (self.GRAD, self.DIRECT):
            _type = weight_decay_type
            msg = 'Invalid weight_decay_type value: {}'.format(_type)
            raise ValueError(msg)

        defaults = {
            'lr': lr,
            'momentum': momentum,
            'nu': nu,
            'weight_decay': weight_decay,
            'weight_decay_type': weight_decay_type,
        }
        super(QHM, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr, nu, momentum = group['lr'], group['nu'], group['momentum']
            weight_decay, weight_decay_type = (
                group['weight_decay'],
                group['weight_decay_type'],
            )

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                param_state = self.state[p]

                if weight_decay != 0:
                    if weight_decay_type == self.GRAD:
                        d_p.add_(p.data, alpha=weight_decay)
                    else:
                        p.data.mul_(1.0 - lr * weight_decay)

                if len(param_state) == 0:
                    param_state['momentum_buffer'] = torch.zeros_like(p.data)

                momentum_buffer = param_state['momentum_buffer']
                momentum_buffer.mul_(momentum).add_(d_p, alpha=1.0 - momentum)

                p.data.add_(momentum_buffer, alpha=-lr * nu)
                p.data.add_(d_p, alpha=-lr * (1.0 - nu))

        return loss


from moai.parameters.optimization.schedule.noop import NoOp
from moai.parameters.optimization.schedule.single import Scheduler as Single
from moai.parameters.optimization.schedule.cascade import Schedulers as Cascade

__all__ = [
    "NoOp",
    "Single",
    "Cascade",
]


from asyncio.windows_events import NULL

import os
import sys
import torch
import glob
import typing
import logging
import functools
import toolz
import json
from collections import namedtuple

import moai.data.datasets.hand.pose.HanCo.importers as importers

__all__ = [
    "HanCoLoader",
]

logger = logging.getLogger(__name__)

class HanCo(torch.utils.data.Dataset):
    __LOADERS__ = {
        'color': importers.loaders.load_color,
        'mask_fg': importers.loaders.load_fg_mask,
        'mask_hand': importers.loaders.load_hand_mask,
        'keypoints3D': importers.loaders.load_keypoints3D,
        'keypoints2D': importers.loaders.load_keypoints2D,
        'mano': importers.loaders.load_mano,
    }

    def __init__(self,
        root:               str,
        data_types:         str,
    ):
        super(HanCo, self).__init__()
        self.splits = ['train', 'val']
        self.filenames = self._get_filenames(root, self.splits)
        self.loaders = [
            functools.partial(
                HanCo.__LOADERS__[type]
            ) for type in data_types
        ]

    def _get_filenames(self,
        root: str,
        splits: typing.List[str]
    ) -> typing.Dict[str, typing.List[str]]:
        if not os.path.exists(root):
            logger.error((
                f"Invalid HanCo root folder ({root}), "
            ))
            sys.exit(2)

        base_filenames = glob.glob(os.path.join(root, 'shape', '**', '*.json'))

        rgb_base = [filename.replace('shape', 'rgb') for filename in base_filenames]
        rgb_base = [filename.replace('json', 'jpg') for filename in rgb_base]
        rgb_filenames = []
        for filename in rgb_base: 
            last = filename.split('\\')[-1]
            for viewpoint in range(8):
                new_name = filename.replace(last, os.path.join('cam' + str(viewpoint), last))
                rgb_filenames.append(new_name)

        shape_filenames = [filename.replace('rgb', 'shape') for filename in rgb_filenames] #this referes to the per image shape
        shape_filenames = [filename.replace('jpg', 'json') for filename in shape_filenames]
        mask_fg_filenames = [filename.replace('rgb', 'mask_fg') for filename in rgb_filenames]
        mask_hand_filenames = [filename.replace('rgb', 'mask_hand') for filename in rgb_filenames]
        xyz_filenames = [filename.replace('shape', 'xyz') for filename in base_filenames]
        calib_filenames = [filename.replace('shape', 'calib') for filename in base_filenames]

        return {
            'rgb_filenames': rgb_filenames,
            'mask_hand_filenames': mask_hand_filenames,
            'mask_fg_filenames': mask_fg_filenames,
            'mano_filenames': shape_filenames,
            'keypoints_filenames': xyz_filenames,
            'calib_filenames': calib_filenames
        }

    def __len__(self) -> int:
        return len(self.filenames['rgb_filenames'])

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        out = { }
        for loader in self.loaders:
            out = toolz.merge(out, loader(self.filenames, index))
        return out


import moai.nn.convolution as mic

import torch

__all__ = [
    "StridedConv2d",
]

class StridedConv2d(torch.nn.Module):  #TODO: Add optional activation as well?
    def __init__(self,
        features: int,
        kernel_size: int=3,
        conv_type: str="conv2d",
        stride: int=2,
        padding: int=1
    ):
        super(StridedConv2d, self).__init__()
        self.conv = mic.make_conv_op(
            kernel_size=kernel_size,
            dilation=1,
            groups=1,
            conv_type=conv_type,
            in_channels=features,
            out_channels=features,
            bias=False,
            stride=2
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


from collections import UserList
from pytorch_lightning import Callback

import torch
import typing
import logging

log = logging.getLogger(__name__)

class ModelCallbacks(UserList):
    def __init__(self,
        list:       typing.Sequence[Callback]=None,
        model:      torch.nn.Module=None,
    ):
        super(ModelCallbacks, self).__init__(list)
        if model:
            self.data.extend((
                c for c in model.children() if isinstance(c, Callback)
            ))


from collections.abc import Callable

import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import logging
import typing

log = logging.getLogger(__name__)

__all__ = ["Collection"]

class Collection(Callable): #TODO: inherit from UserList as well, need to update items to iteration
    def __init__(self,
        items: omegaconf.DictConfig,
        arguments: typing.Sequence[typing.Any]=None,
        name: str="items",
    ):
        self.name = name
        items = [hyu.instantiate(item) for item in items.values()]\
            if not arguments else [
                hyu.instantiate(item, arg) for item, arg in zip(items.values(), arguments)
            ]
        setattr(self, name, items)
        if arguments and len(items) != len(arguments):
            log.warning(f"Inconsistent item ({len(items)}) and argument ({len(arguments)}) count, the matching subset is only used.")

    def items(self) -> typing.Iterable[typing.Any]:
        return getattr(self, self.name)

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:        
        for item in self.items():
            item(tensors)


import torch
import functools

__all__ = [
    "PCKh2D_50",
    "PCKh2D_10"
]

def _calculate_pck(
    gt_kpts:        torch.Tensor,
    pred_kpts:      torch.Tensor,
    gt_masks:       torch.Tensor,
    pred_masks:     torch.Tensor,
    kpts_indices:   list,
    threshold:      float
):
    """
    Detected joint is considered correct if the distance between 
    the predicted and the ground truth joint is within a certain threshold (threshold varies)
    """
    diff = torch.abs(gt_kpts - pred_kpts)
    max_sizes = torch.abs(gt_kpts[:, kpts_indices[0], ...] - gt_kpts[:, kpts_indices[1], ...]).unsqueeze(1)    
    kpts_error_norm = diff / max_sizes
    #count correct threshold
    valid_kpts = torch.zeros_like(kpts_error_norm)
    valid_kpts[kpts_error_norm < threshold] = 1.0
    values2compare = torch.logical_or(pred_masks, gt_masks)
    values2penalize = torch.logical_xor(pred_masks, gt_masks)
    valid_kpts[values2penalize] = 0.0
    pck = 100.0 * torch.mean(valid_kpts[values2compare])
    return pck

class PCK(torch.nn.Module):
    def __init__(self,
        threshold:      float=0.05,
        joint_indices:  list=[0, 2]
    ):
        super(PCK, self).__init__()
        self.threshold = threshold
        self.joint_indices = joint_indices

    def forward(self,
        gt_kpts:            torch.Tensor,
        pred_kpts:          torch.Tensor,
        gt_masks:           torch.Tensor,
        pred_masks:         torch.Tensor,
    ) -> torch.Tensor:
        return _calculate_pck(
            gt_kpts, 
            pred_kpts,
            gt_masks, 
            pred_masks, 
            self.joint_indices, 
            self.threshold
        )

PCKh2D_50 = functools.partial(
    PCK, 
    threshold=0.5,
    joint_indices=[2, 3]
)

PCKh2D_10 = functools.partial(
    PCK, 
    threshold=0.1,
    joint_indices=[2, 3]
)



import torch

import logging

log = logging.getLogger(__name__)

__all__ = ["Default"]

class Default(object):
    def __init__(self, **kwargs):
        pass

    def __call__(self, model: torch.nn.Module) -> None:
        pass


from moai.engine.modules.seed import ManualSeed
from moai.engine.modules.anomaly_detection import AnomalyDetection
from moai.engine.modules.nvidia import (
    NvidiaSMI,
    DebugCUDA
)
from moai.engine.modules.importer import Import
from moai.engine.modules.empty_cache import EmptyCache

__all__ = [
    "AnomalyDetection",
    "DebugCUDA",
    "EmptyCache",
    "Import",
    "NvidiaSMI",
    "Seed",
]


from moai import __version__ as miV
from moai.utils.engine import NoOp as NoInterval
from moai.parameters.optimization import NoOp as NoOptimization
from moai.parameters.optimization.schedule import NoOp as NoScheduling
from moai.parameters.initialization import Default as NoInit
from moai.validation import (
    NoOp as NoValidation,
    Collection as DefaultValidation,
)
from moai.supervision import (
    NoOp as NoSupervision,
    Weighted as DefaultSupervision,
)
from moai.data.iterator import Indexed

import torch
import numpy
import pytorch_lightning
import omegaconf.omegaconf 
import hydra.utils as hyu
import typing
import types
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ['FeedForward']

def _create_supervision_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty supervision block in feedforward model.")
    if not cfg:
        return NoSupervision()
    if '_target_' not in cfg:
        return DefaultSupervision(**cfg)
    else:
        return hyu.instantiate(cfg)

def _create_validation_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty validation block in feedforward model.")    
    if not cfg:
        return NoValidation()
    if '_target_' not in cfg:
        return DefaultValidation(**cfg)
    else:
        return hyu.instantiate(cfg)

def _create_processing_block(
    cfg: omegaconf.DictConfig, 
    attribute: str, 
    monads: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg and attribute in cfg:
        log.warning(f"Empty processing block ({attribute}) in feedforward model.")
    return hyu.instantiate(getattr(cfg, attribute), monads)\
        if cfg and attribute in cfg else torch.nn.Identity()

def _create_interval_block(
    cfg: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg:
        log.warning("Empty interval block in feedforward model.")
    return hyu.instantiate(cfg) if cfg else NoInterval()

def _create_optimization_block(
    cfg: omegaconf.DictConfig,
    params: typing.Union[typing.Iterable[torch.Tensor], typing.Dict[str, torch.Tensor]],
    force: bool=False
):
    if force and not cfg:
        log.warning("No optimizer in feedforward model.")
    return hyu.instantiate(
        # cfg,
        #TODO: hack to get resolved values, should update when omegaconf 2.1 is out and explicit resolve is available
        omegaconf.OmegaConf.create(omegaconf.OmegaConf.to_container(cfg, resolve=True)),
        params
    ) if cfg else NoOptimization(params)

def _create_scheduling_block(
    cfg: omegaconf.DictConfig, 
    optimizers: typing.Sequence[torch.optim.Optimizer],
    force: bool=False
):
    if force and not cfg:
        log.warning("No scheduling used in feedforward model.")
    return hyu.instantiate(cfg, optimizers) if cfg else NoScheduling(optimizers)

def _assign_data(cfg: omegaconf.DictConfig) -> omegaconf.DictConfig:
    has_train = hasattr(cfg, 'train')
    has_test = hasattr(cfg, 'test')
    has_val = hasattr(cfg, 'val')
    if not has_train and not has_test and not has_val:
        log.error("No data have been included in the configuration. Please add the necessary \'- data/[split]/dataset/*: *\' entries.")
    if not has_test and has_val:
        test = omegaconf.OmegaConf.create({'test': cfg.val})
        omegaconf.OmegaConf.set_struct(cfg, False)
        cfg.merge_with(test)
        omegaconf.OmegaConf.set_struct(cfg, True)
        log.warning("No test dataset has been defined, using the validation dataset as a test dataset.")
    if not has_val and has_test:
        val = omegaconf.OmegaConf.create({'val': cfg.test})
        omegaconf.OmegaConf.set_struct(cfg, False)
        cfg.merge_with(val)
        omegaconf.OmegaConf.set_struct(cfg, True)
        log.warning("No validation dataset has been defined, using the test dataset as a validation dataset.")
    return cfg

class FeedForward(pytorch_lightning.LightningModule):
    def __init__(self, 
        data:               omegaconf.DictConfig=None,
        parameters:         omegaconf.DictConfig=None,
        feedforward:        omegaconf.DictConfig=None,
        monads:             omegaconf.DictConfig=None,        
        supervision:        omegaconf.DictConfig=None,
        validation:         omegaconf.DictConfig=None,        
        visualization:      omegaconf.DictConfig=None,
        export:             omegaconf.DictConfig=None,
        hyperparameters:    typing.Union[omegaconf.DictConfig, typing.Mapping[str, typing.Any]]=None,
    ):
        super(FeedForward, self).__init__()        
        self.data = _assign_data(data)
        self.initializer = parameters.initialization if parameters is not None else None
        self.optimization_config = parameters.optimization if parameters is not None else None
        self.schedule_config = parameters.schedule if parameters is not None else None
        self.supervision = _create_supervision_block(supervision)
        self.validation = _create_validation_block(validation) #TODO: change this, "empty processing block" is confusing
        self.preprocess = _create_processing_block(feedforward, "preprocess", monads=monads)
        self.postprocess = _create_processing_block(feedforward, "postprocess", monads=monads)
        self.visualization = _create_interval_block(visualization)
        self.exporter = _create_interval_block(export)        
        #NOTE: __NEEDED__ for loading checkpoint
        hparams = hyperparameters if hyperparameters is not None else { }
        hparams.update({'moai_version': miV})
        #NOTE: @PTL1.5 self.hparams =  hparams
        self.hparams.update(hparams)
        self.global_test_step = 0

    def initialize_parameters(self) -> None:
        init = hyu.instantiate(self.initializer) if self.initializer else NoInit()
        init(self)

    def forward(self,
        tensors:                typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        pass

    def training_step(self, 
        batch:                  typing.Dict[str, torch.Tensor],
        batch_idx:              int,
        optimizer_idx:          int=None,
    ) -> typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        postprocessed = self.postprocess(prediction)
        total_loss, losses = self.supervision(postprocessed)
        #TODO: should add loss maps as return type to be able to forward them for visualization
        losses = toolz.keymap(lambda k: f"train_{k}", losses)
        losses.update({'total_loss': total_loss})        
        self.log_dict(losses, prog_bar=False, logger=True)        
        return { 'loss': total_loss, 'tensors': postprocessed }

    def training_step_end(self, 
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        if self.global_step and (self.global_step % self.visualization.interval == 0):
            self.visualization(train_outputs['tensors'])
        if self.global_step and (self.global_step % self.exporter.interval == 0):
            self.exporter(train_outputs['tensors'])
        return train_outputs['loss']

    def validation_step(self,
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:        
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        #TODO: consider adding loss maps in the tensor dict
        metrics = self.validation(outputs)
        return metrics

    def validation_epoch_end(self,
        outputs: typing.List[dict]
    ) -> None:
        keys = next(iter(outputs), { }).keys()        
        metrics = { }
        for key in keys:
            metrics[key] = numpy.mean(numpy.array(
                [d[key].item() for d in outputs if key in d]
            ))        
        self.log_dict(metrics, prog_bar=True, logger=False, on_epoch=True, sync_dist=True)
        log_metrics = toolz.keymap(lambda k: f"val_{k}", metrics)
        self.log_dict(log_metrics, prog_bar=False, logger=True, on_epoch=True, sync_dist=True)
    
    def test_step(self, 
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        metrics = self.validation(outputs)
        self.global_test_step += 1
        log_metrics = toolz.keymap(lambda k: f"test_{k}", metrics)
        self.log_dict(log_metrics, prog_bar=False, logger=True, on_step=True, on_epoch=False, sync_dist=True)
        return metrics, outputs

    def test_step_end(self,
        metrics_tensors: typing.Tuple[typing.Dict[str, torch.Tensor], typing.Dict[str, torch.Tensor]],        
    ) -> None:
        metrics, tensors = metrics_tensors
        if self.global_test_step and (self.global_test_step % self.exporter.interval == 0):
            self.exporter(tensors)
        if self.global_test_step and (self.global_test_step % self.visualization.interval == 0):
            self.visualization(tensors)
        return metrics

    def test_epoch_end(self, 
        outputs: typing.List[dict]
    ) -> dict:
        keys = next(iter(outputs), { }).keys()
        metrics = { }
        for key in keys:
            metrics[key] = numpy.mean(numpy.array(
                [d[key].item() for d in outputs]
            ))   
        self.log_dict(metrics, prog_bar=False, logger=True, on_epoch=True, sync_dist=True)

    def configure_optimizers(self) -> typing.Tuple[typing.List[torch.optim.Optimizer], typing.List[torch.optim.lr_scheduler._LRScheduler]]:
        log.info(f"Configuring optimizer and scheduler")
        self.optimization = _create_optimization_block(self.optimization_config, self.parameters())
        self.schedule = _create_scheduling_block(self.schedule_config, self.optimization.optimizers)
        return self.optimization.optimizers, self.schedule.schedulers

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.train.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.train.iterator._target_.split('.')[-1]}) train set data iterator")
            train_iterator = hyu.instantiate(self.data.train.iterator)
        else:
            train_iterator = Indexed(
                self.data.train.iterator.datasets,
                self.data.train.iterator.augmentation if hasattr(self.data.train.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.train, 'loader'):
            log.error("Train data loader missing. Please add a data loader (i.e. \'- data/train/loader: torch\') entry in the configuration.")
        else:
            train_loader = hyu.instantiate(self.data.train.loader, train_iterator)
        return train_loader

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.val.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.val.iterator._target_.split('.')[-1]}) validation set data iterator")
            val_iterator = hyu.instantiate(self.data.val.iterator)
        else:
            val_iterator = Indexed(
                self.data.val.iterator.datasets,
                self.data.val.iterator.augmentation if hasattr(self.data.val.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.val, 'loader'):
            log.error("Validation data loader missing. Please add a data loader (i.e. \'- data/val/loader: torch\') entry in the configuration.")
        else:
            validation_loader = hyu.instantiate(self.data.val.loader, val_iterator)
        return validation_loader

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.test.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.test.iterator._target_.split('.')[-1]}) test set data iterator")
            test_iterator = hyu.instantiate(self.data.test.iterator)
        else:
            test_iterator = Indexed(
                self.data.test.iterator.datasets,
                self.data.test.iterator.augmentation if hasattr(self.data.test.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.test, 'loader'):
            log.error("Test data loader missing. Please add a data loader (i.e. \'- data/test/loader: torch\') entry in the configuration.")
        else:
            test_loader = hyu.instantiate(self.data.test.loader, test_iterator)
        return test_loader



import albumentations.augmentations.transforms as aug

__all__ = ["RandomGamma"]

#NOTE: wrapper as to circumvent hydra/tuple issue
class RandomGamma(aug.RandomGamma): 
    def __init__(self,
        low: float=80,
        high: float=120,
        eps: float=1e-7,
        p: float=0.5
    ):
        super(RandomGamma, self).__init__(
            gamma_limit=(low, high),
            eps=eps,
            always_apply=False, 
            p=p
        )


import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import toolz

__all__ = ["Zipped"]

logger = logging.getLogger(__name__)

class Zipper(torch.utils.data.Dataset):
    def __init__(self,
        datasets = typing.Sequence[torch.utils.data.Dataset],
    ):
        self.datasets = datasets
        if len(list(toolz.unique(map(len, self.datasets)))) > 1:
            def _name_key(d: torch.utils.data.Dataset):
                return f"{d.__class__.__name__}: {len(d)}"
            logger.warning(
                f"Zipped datasets are of unequal lengths ("
                f"{map(_name_key, self.datasets)}),"
                f" will reduce length to smallest one ({min(map(len, self.datasets))})."
            )

    def __len__(self):
        return min(map(len, self.datasets))
    
    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        out = { }
        for d in self.datasets:
            out = toolz.merge(out, d[index])
        return out

class Zipped(torch.utils.data.Dataset):    
    def __init__(self,
        datasets: omegaconf.DictConfig,
        augmentation: omegaconf.DictConfig=None,
    ):
        super(Zipped, self).__init__()
        if augmentation is not None:
            self.dataset = hyu.instantiate(
                augmentation,
                Zipper([hyu.instantiate(d) for d in datasets.values()])
            )
        else:
            from moai.data.augmentation import NoOp
            self.dataset = NoOp(Zipper([hyu.instantiate(d) for d in datasets.values()]))

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return self.dataset[index]


import moai.networks.lightning as minet
import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.sampling.spatial.downsample as mids
import moai.modules.lightning as mimod
import moai.nn.utils as miu

import torch

import hydra.utils as hyu
import omegaconf.omegaconf as omegaconf
import typing
import logging

log = logging.getLogger(__name__)

#NOTE: from https://github.com/anibali/pytorch-stacked-hourglass/blob/master/src/stacked_hourglass/model.py
#NOTE: from https://github.com/princeton-vl/pytorch_stacked_hourglass/blob/master/models/posenet.py

__all__ = ["StackedHourglass"]

class StackedHourglass(minet.FeedForward):
    def __init__(self,
        configuration:  omegaconf.DictConfig,
        modules:        omegaconf.DictConfig,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(StackedHourglass, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            export=export, visualization=visualization,            
        )        
        self.stacks = configuration.stacks
        preproc = configuration.preproc
        projection = configuration.projection
        prediction = configuration.prediction
        merge = configuration.merge
        hourglass = modules['hourglass']
        self.pre = torch.nn.Sequential(
            mic.make_conv_block(
                block_type=preproc.block,
                convolution_type=preproc.convolution,
                in_features=configuration.in_features, 
                out_features=hourglass.features // 4,
                activation_type=preproc.activation,
                convolution_params={
                    "kernel_size": preproc.stem.kernel_size,
                    "stride": preproc.stem.stride,
                    "padding": preproc.stem.padding,
                },
            ),
            mires.make_residual_block(
                block_type=preproc.residual,
                convolution_type=preproc.convolution,
                in_features=hourglass.features // 4,
                out_features=hourglass.features // 2,
                bottleneck_features=hourglass.features // 2,
                activation_type=preproc.activation,
                strided=False,
            ),
            mids.make_downsample(
                downscale_type=preproc.downscale,
                features=hourglass.features // 2,
                kernel_size=3 if preproc.downscale == 'maxpool2d_aa' else 2,
            ),
            mires.make_residual_block(
                block_type=preproc.residual,
                convolution_type=preproc.convolution,
                in_features=hourglass.features // 2,
                out_features=hourglass.features // 2,
                bottleneck_features=hourglass.features // 2,
                activation_type=preproc.activation,
                strided=False,
            ),
            mires.make_residual_block(
                block_type=preproc.residual,
                convolution_type=preproc.convolution,
                in_features=hourglass.features // 2,
                out_features=hourglass.features,
                bottleneck_features=hourglass.features,
                activation_type=preproc.activation,
                strided=False,
            ),
        )
        
        self.hgs = torch.nn.ModuleList([
                torch.nn.Sequential(
                    hyu.instantiate(hourglass)
                ) for i in range(self.stacks)
            ] 
        )
        
        self.features = torch.nn.ModuleList([
                torch.nn.Sequential(
                    mires.make_residual_block(
                        block_type=preproc.residual,
                        convolution_type=preproc.convolution,
                        in_features=hourglass.features,
                        out_features=hourglass.features,
                        bottleneck_features=hourglass.features,
                        activation_type=preproc.activation,
                        strided=False,
                    ),
                    mic.make_conv_block(
                        block_type=projection.block,
                        convolution_type=projection.convolution,
                        in_features=hourglass.features, 
                        out_features=hourglass.features,
                        activation_type=projection.activation,
                        convolution_params={"kernel_size": 1, "padding": 0},
                    )
                ) for i in range(self.stacks)
            ]
        )
        
        self.outs = torch.nn.ModuleList([
            mic.make_conv_block(
                block_type=prediction.block,
                convolution_type=prediction.convolution,
                in_features=hourglass.features, 
                out_features=configuration.out_features,
                activation_type=prediction.activation,
                convolution_params={
                    "kernel_size": 1, "padding": 0, 
                },
                activation_params={"inplace": True} 
            ) for i in range(self.stacks)         
        ])
        self.merge_features = torch.nn.ModuleList([
            torch.nn.Sequential(
                mic.make_conv_1x1(
                    convolution_type=projection.convolution,
                    in_channels=hourglass.features,
                    out_channels=hourglass.features
                ),
                torch.nn.Dropout2d(p=merge.dropout, inplace=True)\
                    if merge.dropout > 0.0 else torch.nn.Identity()
            ) for i in range(self.stacks-1)
        ])
        self.merge_preds = torch.nn.ModuleList([
            torch.nn.Sequential(
                mic.make_conv_1x1(
                    convolution_type=projection.convolution,
                    in_channels=configuration.out_features,
                    out_channels=hourglass.features
                ),
                torch.nn.Dropout2d(p=prediction.dropout, inplace=False)\
                    if prediction.dropout > 0.0 else torch.nn.Identity()
            ) for i in range(self.stacks-1)
        ])
        self.input = configuration.input
        self.output_prefix = configuration.output

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:        
        img = td[self.input]
        x = self.pre(img)
        combined_hm_preds = []
        for i in range(self.stacks):
            hg = self.hgs[i](x)
            feature = self.features[i](hg)
            preds = self.outs[i](feature)
            combined_hm_preds.append(preds)
            if i < self.stacks - 1:                
                x = x + self.merge_preds[i](preds) + self.merge_features[i](feature)
        for i, heatmap in enumerate(combined_hm_preds):
            td[f"{self.output_prefix}_{i+1}"] = heatmap
        return td

    


import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["Threshold"]

class Threshold(torch.nn.Module):
    CAST_OPS = {
        "float": lambda t: t.float(),
        "byte": lambda t: t.byte(),
    }

    def __init__(self,
        value: float,
        comparison: str="lower",
        dtype: str="float", # "byte" #TODO: change to 'mask' and 'weights' semantics?
    ):
        super(Threshold, self).__init__()
        self.threshold = value
        self.comp_op = torch.le if comparison == "lower"\
            else (
                torch.ge if comparison == "greater" else torch.ge                    
            ) #TODO: update properly (avoid implicit resolving to ge for all cases)
        if dtype not in Threshold.CAST_OPS:
            log.error("Casting operation type for Threshold monad should be either float or byte")
        self.cast_op = Threshold.CAST_OPS[dtype]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.cast_op(self.comp_op(x, self.threshold))


from collections.abc import Callable

import visdom
import typing
import torch

__all__ = ["Base"]

class Base(Callable):
    def __init__(self,
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097, #NOTE: visdom's default port
    ):
        self.env_name, self.ip, self.port, self.viz = \
            name, ip, port, None

    @property
    def visualizer(self):
        if self.viz is not None:
            return self.viz

        self.viz = visdom.Visdom(
            server=self.ip,
            port=self.port,
            env=self.env_name,
            use_incoming_socket=False
        )
        return self.viz

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        raise NotImplementedError()


import moai.utils.engine as mieng

import sys
import torch
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["NoOp"]

class NoOp(mieng.Interval):
    def __init__(self, 
        *args:      typing.Sequence[typing.Any],
        **kwargs:   typing.Mapping[str, typing.Any]
    ):
        super(NoOp, self).__init__(sys.maxsize * 2 + 1) # big enough interval

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        pass


from moai.parameters.optimization.optimizers.lookahead import Lookahead

import math
import torch
import typing

__all__ = ['Ranger']

#NOTE: from https://github.com/mgrankin/over9000/blob/master/ranger.py

class Ranger(Lookahead):
     """Implements a combination of RAdam and Lookahead:

     - [ON THE VARIANCE OF THE ADAPTIVE LEARNING RATE AND BEYOND](https://arxiv.org/pdf/1908.03265.pdf)
     - [Lookahead Optimizer: k steps forward, 1 step back](https://arxiv.org/pdf/1907.08610.pdf)
     - **Implementation**: [GitHub @ mgrankin](https://github.com/mgrankin/over9000)

     """
     def __init__(self,
          params: typing.Iterator[torch.nn.Parameter],
          alpha: float=0.5,
          k: int=5,
          lr: float=1e-3,
          betas: typing.Tuple[float, float]=(0.9, 0.999),
          eps: float=1e-8,
          weight_decay: float=0,
          degenerated_to_sgd: bool=True,
     ):
          super(Ranger, self).__init__(
               mio.RAdam(params, 
                    lr=lr, betas=betas, eps=eps,
                    weight_decay=weight_decay,
                    degenerated_to_sgd=degenerated_to_sgd,
               ), k=k, alpha=alpha
          )


import math
import torch
import typing

__all__ = ['NovoGrad']

#NOTE: modified from https://github.com/lonePatient/NovoGrad-pytorch

class NovoGrad(torch.optim.Optimizer):
    """Implements NovoGrad algorithm.

    - **Paper**: [Training Deep Networks with Stochastic Gradient Normalized by Layerwise Adaptive Second Moments](https://arxiv.org/pdf/1905.11286.pdf)
    - **Implementation**: [GitHub @ lonePatient](https://github.com/lonePatient/NovoGrad-pytorch)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-2)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.95, 0.98))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
    Example:
        >>> model = ResNet()
        >>> optimizer = NovoGrad(model.parameters(), lr=1e-2, weight_decay=1e-5)
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=0.01,
        betas: typing.Tuple[float, float]=(0.95, 0.98),
        eps: float=1e-8,
        weight_decay: float=0,
        grad_averaging: bool=False
    ):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps,weight_decay=weight_decay,grad_averaging = grad_averaging)
        super().__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        loss = None
        if closure is not None:
            loss = closure()
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('NovoGrad does not support sparse gradients')
                state = self.state[p]
                g_2 = torch.sum(grad ** 2)
                if len(state) == 0:
                    state['step'] = 0
                    state['moments'] = grad.div(g_2.sqrt() +group['eps']) + \
                                       group['weight_decay'] * p.data
                    state['grads_ema'] = g_2
                moments = state['moments']
                grads_ema = state['grads_ema']
                beta1, beta2 = group['betas']
                state['step'] += 1
                grads_ema.mul_(beta2).add_(1 - beta2, g_2)

                denom = grads_ema.sqrt().add_(group['eps'])
                grad.div_(denom)
                # weight decay
                if group['weight_decay'] != 0:
                    decayed_weights = torch.mul(p.data, group['weight_decay'])
                    grad.add_(decayed_weights)

                # Momentum --> SAG
                if group['grad_averaging']:
                    grad.mul_(1.0 - beta1)

                moments.mul_(beta1).add_(grad) # velocity

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1
                p.data.add_(-step_size, moments)

        return loss

class AdamW(torch.optim.Optimizer):
    """Implements Adam algorithm (with proper weight decay).

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
    Example:
        >>> model = alexnet()
        >>> optimizer = AdamW(model.parameters(), lr=1e-3, weight_decay=1e-5)
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter], 
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        amsgrad: bool=False
    ):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps,weight_decay=weight_decay, amsgrad=amsgrad)
        #super(AdamW, self).__init__(params, defaults)
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']
                state['step'] += 1
                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group['eps'])
                else:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1
                if group['weight_decay'] != 0:
                    decayed_weights = torch.mul(p.data, group['weight_decay'])
                    p.data.addcdiv_(-step_size, exp_avg, denom)
                    p.data.sub_(decayed_weights)
                else:
                    p.data.addcdiv_(-step_size, exp_avg, denom)
        return loss


import moai.nn.convolution as mic
import moai.nn.activation as mia

import torch

__all__ = [
    "Conv2dBlock",
]

class Conv2dBlock(torch.nn.Module):
    def __init__(self,
        convolution_type:str,
        activation_type: str,
        in_features: int,
        out_features: int,
        convolution_params: dict,
        activation_params: dict,
        **kwargs: dict
    ):
        super(Conv2dBlock, self).__init__()
        self.conv = mic.make_conv_op(
            convolution_type=convolution_type,
            in_channels=in_features,
            out_channels=out_features,
            **convolution_params #TODO: either merge kwargs here or in the factory method
        )
        self.activation = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.conv(x))


from moai.networks.lightning.feedforward import FeedForward
from moai.networks.lightning.wrapper import Wrapper
from moai.networks.lightning.optimizer import Optimizer
from moai.networks.lightning.none import Null

__all__ = [
    "Wrapper",
    "FeedForward",
    "Optimizer",
    "Null",
]


import torch

class L2(torch.nn.Module):
    def __init__(self):
        super(L2, self).__init__()

    def forward(self,        
        pred:       torch.Tensor,
        gt:         torch.Tensor=None, #NOTE: w/o gt it serves as a prior
        weights:    torch.Tensor=None, # float tensor
        mask:       torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        l2 = ((gt - pred) ** 2) if gt is not None else pred ** 2
        if weights is not None:
            l2 = l2 * weights
        if mask is not None:
            l2 = l2[mask]        
        return l2


import torch

#NOTE: modified code from https://github.com/NVIDIAGameWorks/kaolin/blob/master/kaolin/rep/Mesh.py

__all__ = ["Mesh", "TriangleMesh"]

class Mesh():
    """ Abstract class to represent 3D polygon meshes. """

    def __init__(self, vertices: torch.Tensor, faces: torch.Tensor,
                 uvs: torch.Tensor, face_textures: torch.Tensor,
                 textures: torch.Tensor, edges: torch.Tensor, edge2key: dict, vv: torch.Tensor,
                 vv_count: torch.Tensor, vf: torch.Tensor, vf_count: torch.Tensor,
                 ve: torch.Tensor, ve_count: torch.Tensor, ff: torch.Tensor,
                 ff_count: torch.Tensor, ef: torch.Tensor, ef_count: torch.Tensor,
                 ee: torch.Tensor, ee_count: torch.Tensor):\

        # Vertices of the mesh
        self.vertices = vertices
        # Faces of the mesh
        self.faces = faces
        # uv coordinates of each vertex
        self.uvs = uvs
        # uv indecies for each face
        self.face_textures = face_textures
        # texture for each face
        self.textures = textures
        # Edges of the mesh
        self.edges = edges
        # Dictionary that maps an edge (tuple) to an edge idx
        self.edge2key = edge2key
        # Vertex-Vertex neighborhood tensor (for each vertex, contains
        # indices of the vertices neighboring it)
        self.vv = vv
        # Number of vertices neighbouring each vertex
        self.vv_count = vv_count
        # Vertex-Face neighborhood tensor
        self.vf = vf
        # Number of faces neighbouring each vertex
        self.vf_count = vf_count
        # Vertex-Edge neighborhood tensor
        self.ve = ve
        # Number of edges neighboring each vertex
        self.ve_count = ve_count
        # Face-Face neighborhood tensor
        self.ff = ff
        # Number of faces neighbouring each face
        self.ff_count = ff_count
        # Edge-Face neighbourhood tensor
        self.ef = ef
        # Number of edges neighbouring each face
        self.ef_count = ef_count
        # Edge-Edge neighbourhood tensor
        self.ee = ee
        # Number of edges neighbouring each edge
        self.ee_count = ee_count
        # adjacency matrix for verts
        self.adj = None
    
    @classmethod
    def from_obj(self, filename: str, with_vt: bool = False,
                 enable_adjacency: bool = False, texture_res=4):
        r"""Loads object in .obj wavefront format.
        Args:
            filename (str) : location of file.
            with_vt (bool): objects loaded with textures specified by vertex
                textures.
            enable_adjacency (bool): adjacency information is computed.
            texture_res (int): resolution of loaded face colors.
        Note: the with_vt parameter requires cuda.
        Example:
            >>> mesh = Mesh.from_obj('model.obj')
            >>> mesh.vertices.shape
            torch.Size([482, 3])
            >>> mesh.faces.shape
            torch.Size([960, 3])
        """
        # run through obj file and extract obj info
        vertices = []
        faces = []
        face_textures = []
        uvs = []
        with open(filename, 'r') as mesh:
            for line in mesh:
                data = line.split()
                if len(data) == 0:
                    continue
                if data[0] == 'v':
                    vertices.append(data[1:])
                elif data[0] == 'vt':
                    uvs.append(data[1:3])
                elif data[0] == 'f':
                    if '//' in data[1]:
                        data = [da.split('//') for da in data]
                        faces.append([int(d[0]) for d in data[1:]])
                        face_textures.append([int(d[1]) for d in data[1:]])
                    elif '/' in data[1]:
                        data = [da.split('/') for da in data]
                        faces.append([int(d[0]) for d in data[1:]])
                        face_textures.append([int(d[1]) for d in data[1:]])
                    else:
                        faces.append([int(d) for d in data[1:]])
                        continue
        vertices = torch.FloatTensor([float(el) for sublist in vertices for el in sublist]).view(-1, 3)
        faces = torch.LongTensor(faces) - 1

        # compute texture info
        textures = None
        if with_vt:
            with open(filename, 'r') as f:
                textures = None
                for line in f:
                    if line.startswith('mtllib'):
                        filename_mtl = os.path.join(
                            os.path.dirname(filename), line.split()[1])
                        textures = self.load_textures(
                            filename, filename_mtl, texture_res)

                f.close()

        if len(uvs) > 0:
            uvs = torch.FloatTensor([float(el) for sublist in uvs for el in sublist]).view(-1, 2)
        else:
            uvs = None
        if len(face_textures) > 0:
            face_textures = torch.LongTensor(face_textures) - 1
        else:
            face_textures = None

        if enable_adjacency:
            edge2key, edges, vv, vv_count, ve, ve_count, vf, vf_count, ff, ff_count, \
                ee, ee_count, ef, ef_count = self.compute_adjacency_info(
                    vertices, faces)
        else:
            edge2key, edges, vv, vv_count, ve, ve_count, vf, vf_count, ff, \
                ff_count, ee, ee_count, ef, ef_count = None, None, None, \
                None, None, None, None, None, None, None, None, None, None, \
                None

        output = self(vertices, faces, uvs, face_textures, textures, edges,
                    edge2key, vv, vv_count, vf, vf_count, ve, ve_count, ff, ff_count,
                    ef, ef_count, ee, ee_count)
        
        return output


class TriangleMesh(Mesh):
    """ Abstract class to represent 3D Trianlge meshes. """

    def __init__(self, vertices: torch.Tensor, faces: torch.Tensor,
                 uvs: torch.Tensor, face_textures: torch.Tensor,
                 textures: torch.Tensor, edges: torch.Tensor, edge2key: dict,
                 vv: torch.Tensor, vv_count: torch.Tensor,
                 vf: torch.Tensor, vf_count: torch.Tensor,
                 ve: torch.Tensor, ve_count: torch.Tensor,
                 ff: torch.Tensor, ff_count: torch.Tensor,
                 ef: torch.Tensor, ef_count: torch.Tensor,
                 ee: torch.Tensor, ee_count: torch.Tensor):

        # Vertices of the mesh
        self.vertices = vertices
        # Faces of the mesh
        self.faces = faces
        # uv coordinates of each vertex
        self.uvs = uvs
        # uv indecies for each face
        self.face_textures = face_textures
        # texture for each face
        self.textures = textures
        # Edges of the mesh
        self.edges = edges
        # Dictionary that maps an edge (tuple) to an edge idx
        self.edge2key = edge2key
        # Vertex-Vertex neighborhood tensor (for each vertex, contains
        # indices of the vertices neighboring it)
        self.vv = vv
        # Number of vertices neighbouring each vertex
        self.vv_count = vv_count
        # Vertex-Face neighborhood tensor
        self.vf = vf
        # Number of faces neighbouring each vertex
        self.vf_count = vf_count
        # Vertex-Edge neighborhood tensor
        self.ve = ve
        # Number of edges neighboring each vertex
        self.ve_count = ve_count
        # Face-Face neighborhood tensor
        self.ff = ff
        # Number of faces neighbouring each face
        self.ff_count = ff_count
        # Edge-Face neighbourhood tensor
        self.ef = ef
        # Number of edges neighbouring each face
        self.ef_count = ef_count
        # Edge-Edge neighbourhood tensor
        self.ee = ee
        # Number of edges neighbouring each edge
        self.ee_count = ee_count
        # adjacency matrix for verts
        self.adj = None





from moai.utils.arguments import assert_positive
import torch
import typing
import logging

logger = logging.getLogger(__name__)

__all__ = ["WeakPerspective"]

class WeakPerspective(torch.nn.Module): #NOTE: fixed focal/principal, optimized rot/trans
    def __init__(self,
        focal_length:       typing.Union[float, typing.Tuple[float, float]]=5000.0,
        principal_point:    typing.Union[float, typing.Tuple[float, float]]=0.5,
        width:              int=None,
        height:             int=None,
        rotation:           torch.Tensor=None,
        translation:        torch.Tensor=None,
        persistent:         bool=False,
    ):
        super(WeakPerspective, self).__init__()
        mat = torch.zeros([1, 2, 2])
        mat[:, 0, 0] = focal_length if isinstance(focal_length, float) else focal_length[0]
        mat[:, 1, 1] = focal_length if isinstance(focal_length, float) else focal_length[1]
        self.register_buffer("mat", mat, persistent=persistent)
        center = torch.zeros([1, 2])
        if principal_point is None:
            assert_positive(logger, "width", width)
            assert_positive(logger, "height", height)
            center[:, 0] = width // 2
            center[:, 1] =  height // 2
        else:
            center[:, 0] = torch.scalar_tensor(
                principal_point if isinstance(principal_point, float) else principal_point[0]
            )
            center[:, 1] = torch.scalar_tensor(
                principal_point if isinstance(principal_point, float) else principal_point[1]
            )
        self.register_buffer('principal_point', center, persistent=persistent)                                                         
        if rotation is None:
             rotation = torch.eye(3).unsqueeze(dim=0)
        rotation = torch.nn.Parameter(rotation, requires_grad=True)
        self.register_parameter('rotation', rotation)
        if translation is None:
            translation = torch.zeros([1, 3])
        translation = torch.nn.Parameter(translation, requires_grad=True)
        self.register_parameter('translation', translation)

    def forward(self, 
        points:             torch.Tensor,
        image:              torch.Tensor=None,
        rotation:           torch.Tensor=None,
        translation:        torch.Tensor=None,
        #TODO: update with focal/principal inputs as well        
    ) -> torch.Tensor:
        if image is not None:
            h, w = image.shape[-2:]
            with torch.no_grad():
                self.principal_point[:, 0] = w // 2
                self.principal_point[:, 1] = h // 2
        R = rotation if rotation is not None else self.rotation.expand(points.shape[0], 3, 3)
        t = translation if translation is not None else self.translation.expand(points.shape[0], 3)
        camera_transform = torch.cat([
            torch.nn.functional.pad(R, [0, 0, 0, 1]),
            torch.nn.functional.pad(t.unsqueeze(dim=-1), [0, 0, 0, 1], value=1)
        ], dim=2)
        z = torch.ones_like(points[..., :1])
        homogeneous_points = torch.cat([points, z], dim=-1)
        projected_points = torch.einsum('bki,bji->bjk',
            [camera_transform, homogeneous_points]
        )
        img_points = torch.div(projected_points[:, :, :2],
            projected_points[:, :, 2].unsqueeze(dim=-1)
        )
        img_points = torch.einsum('bki,bji->bjk', [self.mat, img_points]) \
            + self.principal_point.unsqueeze(dim=1) #TODO: add principal in mat
        return img_points


import torch
import numpy as np
import typing

__all__ = ['UniformSimulatedAnnealing', 'GaussianSimulatedAnnealing']

#NOTE: modified from https://github.com/jramapuram/SimulatedAnnealing

class UniformSampler(object):
    def __init__(self, 
        minval: float,
        maxval: float,
        dtype: str='float',
        cuda: bool=False
    ):
        self.minval = minval
        self.maxval = maxval
        self.cuda = cuda
        self.dtype_str = dtype
        dtypes = {
            'float': torch.cuda.FloatTensor if cuda else torch.FloatTensor,
            'int': torch.cuda.IntTensor if cuda else torch.IntTensor,
            'long': torch.cuda.LongTensor if cuda else torch.LongTensor
        }
        self.dtype = dtypes[dtype]

    def sample(self, size):
        if self.dtype_str == 'float':
            return self.dtype(*size).uniform_(
                self.minval, self.maxval
            )
        elif self.dtype_str == 'int' or self.dtype_str == 'long':
            return self.dtype(*size).random_(
                self.minval, self.maxval + 1
            )
        else:
            raise Exception("unknown dtype")

class GaussianSampler(object):
    def __init__(self, 
        mu: float,
        sigma: float,
        dtype: str='float',
        cuda: bool=False
    ):
        self.sigma = sigma
        self.mu = mu
        self.cuda = cuda
        self.dtype_str = dtype
        dtypes = {
            'float': torch.cuda.FloatTensor if cuda else torch.FloatTensor,
            'int': torch.cuda.IntTensor if cuda else torch.IntTensor,
            'long': torch.cuda.LongTensor if cuda else torch.LongTensor
        }
        self.dtype = dtypes[dtype]

    def sample(self, size):
        ''' pytorch doesnt support int or long normal distrs
            so we will resolve to casting '''
        rand_float = torch.cuda.FloatTensor if self.cuda else torch.FloatTensor
        rand_block = rand_float(*size).normal_(self.mu, self.sigma)

        if self.dtype_str == 'int' or self.dtype_str == 'long':
            rand_block = rand_block.type(self.dtype)

        return rand_block

class SimulatedAnnealing(torch.optim.Optimizer):
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        sampler: typing.Union[UniformSampler, GaussianSampler],
        tau0: float=1.0,
        anneal_rate: float=0.0003,
        min_temp: float=1e-5,
        anneal_every: int=100000,
        hard: bool=False,
        hard_rate: float=0.9
    ):
        defaults = dict(sampler=sampler, tau0=tau0, tau=tau0, anneal_rate=anneal_rate,
                        min_temp=min_temp, anneal_every=anneal_every,
                        hard=hard, hard_rate=hard_rate, iteration=0)
        super(SimulatedAnnealing, self).__init__(params, defaults)


    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        if closure is None:
            raise Exception("loss closure is required to do SA")

        loss = closure()

        for group in self.param_groups:
            # the sampler samples randomness
            # that is used in optimizations
            sampler = group['sampler']

            # clone all of the params to keep in case we need to swap back
            cloned_params = [p.clone() for p in group['params']]

            for p in group['params']:
                # anneal tau if it matches the requirements
                if group['iteration'] > 0 \
                   and group['iteration'] % group['anneal_every'] == 0:
                    if not group['hard']:
                        # smoother annealing: consider using this over hard annealing
                        rate = -group['anneal_rate'] * group['iteration']
                        group['tau'] = np.maximum(group['tau0'] * np.exp(rate),
                                                  group['min_temp'])
                    else:
                        # hard annealing
                        group['tau'] = np.maximum(group['hard_rate'] * group['tau'],
                                                  group['min_temp'])

                random_perturbation = group['sampler'].sample(p.data.size())
                p.data = p.data / torch.norm(p.data)
                p.data.add_(random_perturbation)
                group['iteration'] += 1

            # re-evaluate the loss function with the perturbed params
            # if we didn't accept the new params swap back and return
            loss_perturbed = closure()
            final_loss, is_swapped = self.anneal(loss, loss_perturbed, group['tau'])
            if is_swapped:
                for p, pbkp in zip(group['params'], cloned_params):
                    p.data = pbkp.data

            return final_loss


    def anneal(self, loss, loss_perturbed, tau):
        '''returns loss, is_new_loss'''
        def acceptance_prob(old, new, temp):
            return torch.exp((old - new)/temp)

        if loss_perturbed.data[0] < loss.data[0]:
            return loss_perturbed, True
        else:
            # evaluate the metropolis criterion
            ap = acceptance_prob(loss, loss_perturbed, tau)
            log.info(f"old = {loss.data[0]} | pert = {loss_perturbed.data[0]} | ap = {ap.data[0]} | tau = {tau}")
            if ap.data[0] > np.random.rand():
                return loss_perturbed, True

            # return the original loss if above fails
            # or if the temp is now annealed
            return loss, False

class UniformSimulatedAnnealing(SimulatedAnnealing):
    """Implements Simulated Annealing optimization with uniform sampling.

    - **Paper**: [Optimization by Simulated Annealing](https://www.sciencedirect.com/science/article/pii/B9780080515816500593)
    - **Implementation**: [GitHub @ jramapuram](https://github.com/jramapuram/SimulatedAnnealing)

    """
    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        minval: float,
        maxval: float,
        tau0: float=1.0,
        anneal_rate: float=0.0003,
        min_temp: float=1e-5,
        anneal_every: int=100000,
        hard: bool=False,
        hard_rate: float=0.9,        
        dtype: str='float',
        cuda: bool=False,
    ):
        super(UniformSimulatedAnnealing, self).__init__(params,
            sampler= UniformSampler(
                minval=minval, maxval=maxval, dtype=dtype, cuda=cuda,
            ), 
            tau0=tau0, anneal_rate=anneal_rate, anneal_every=anneal_every,
            min_temp=min_temp, hard=hard, hard_rate=hard_rate
        )

class GaussianSimulatedAnnealing(SimulatedAnnealing):
    """Implements Simulated Annealing optimization with Gaussian sampling.

    - **Paper**: [Optimization by Simulated Annealing](https://www.sciencedirect.com/science/article/pii/B9780080515816500593)
    - **Implementation**: [GitHub @ jramapuram](https://github.com/jramapuram/SimulatedAnnealing)
    
    """
    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        mu: float,
        sigma: float,
        tau0: float=1.0,
        anneal_rate: float=0.0003,
        min_temp: float=1e-5,
        anneal_every: int=100000,
        hard: bool=False,
        hard_rate: float=0.9,        
        dtype: str='float',
        cuda: bool=False,        
    ):
        super(GaussianSimulatedAnnealing, self).__init__(params,
            sampler= GaussianSampler(
                mu=mu, sigma=sigma, dtype=dtype, cuda=cuda,
            ), 
            tau0=tau0, anneal_rate=anneal_rate, anneal_every=anneal_every,
            min_temp=min_temp, hard=hard, hard_rate=hard_rate
        )


import moai.networks.lightning as minet

import torch

import hydra.utils as hyu
import typing
import types
import logging

log = logging.getLogger(__name__)

__all__ = ["Pretrained"]

class Pretrained(object):    
    def __init__(self,         
        filename: str,
        strict: bool=True
    ):
        self.filename, self.strict = filename, strict

    def __call__(self, 
        model: torch.nn.Module
    ) -> None:
        log.info(f"Loading pretrained model weights from {self.filename}")
        checkpoint = torch.load(self.filename,
            map_location=lambda storage, loc: storage
        )
        #TODO: optimizer state + epoch and other metadata 
        if 'hparams' in checkpoint.keys(): #TODO: is 'hyper_parameters' now
            ckpt_hparams = checkpoint['hparams']
            model.hparams = ckpt_hparams \
                if isinstance(ckpt_hparams, types.SimpleNamespace) \
                    else types.SimpleNamespace(**ckpt_hparams)
        if isinstance(model, minet.Wrapper):
            model.load_state_dict(checkpoint['state_dict'], strict=self.strict)
        else:
            model.load_state_dict(checkpoint['state_dict'], strict=self.strict)
            model.on_load_checkpoint(checkpoint) #NOTE: check why this is needed




from moai.monads.utils.common import dim_list

import torch

class SqRel(torch.nn.Module):
    def __init__(self):
        super(SqRel, self).__init__()

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        sqrel = ((gt - pred) ** 2) / gt
        if weights is not None:
            sqrel = sqrel * weights
        if mask is not None:
            sqrel = sqrel[mask]
        if weights is None:
            return torch.mean(torch.mean(sqrel, dim=dim_list(gt)))
        else:
            return torch.mean(
                torch.sum(sqrel, dim=dim_list(gt))
                / torch.sum(weights, dim=dim_list(gt))
            )


import torch
import typing

__all__ = ["Keypoints"]

class Keypoints(torch.nn.Module):
    def __init__(self,
        keypoints: typing.Union[str,typing.List[typing.List[float]]],
    ):
        super(Keypoints,self).__init__()
        #TODO: add numpy loading
        if isinstance(keypoints,str):
            #TODO:load keypoints from file
            pass
        self.register_buffer("keypoints", torch.tensor(keypoints).float())

    def forward(self,x: torch.Tensor) -> torch.Tensor:
        b = x.shape[0]
        return self.keypoints.expand(b,*self.keypoints.shape)


import torch
import typing

class IntegerMadam(torch.optim.Optimizer):
    r"""Implements the Integer Madam optimisation algorithm.

        ??? info "[Learning compositional functions via multiplicative weight updates](https://arxiv.org/pdf/2006.14560.pdf)"
            **Abstract**: Compositionality is a basic structural feature of both biological and artificial
            neural networks. Learning compositional functions via gradient descent incurs
            well known problems like vanishing and exploding gradients, making careful
            learning rate tuning essential for real-world applications. This paper proves that
            multiplicative weight updates satisfy a descent lemma tailored to compositional
            functions. Based on this lemma, we derive Madam—a multiplicative version of
            the Adam optimiser—and show that it can train state of the art neural network
            architectures without learning rate tuning. We further show that Madam is easily
            adapted to train natively compressed neural networks by representing their weights
            in a logarithmic number system. We conclude by drawing connections between
            multiplicative weight updates and recent findings about synapses in biology

        ??? info "[GitHub @ jxbz/madam](https://github.com/jxbz/madam)"

        Arguments:
            base_lr(float, optional): 
            lr_factor(float, optional): 
            levels(int, optional): 
            p_scale(float, optional): 
            g_bound(float, optional):             
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter], 
        base_lr: float=0.001,
        lr_factor: float=10,
        levels: int=4096,
        p_scale: float=3.0,
        g_bound: float=10.0
    ):
        self.p_scale = p_scale
        self.g_bound = g_bound
        defaults = dict(base_lr=base_lr, lr_factor=lr_factor, levels=levels)
        super(IntegerMadam, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                lr = group['base_lr']
                lr_factor = group['lr_factor']
                levels = group['levels']

                state = self.state[p]
                if len(state) == 0:
                    state['max'] = self.p_scale*(p*p).mean().sqrt().item()
                    state['integer'] = torch.randint_like(p, low=0, high=levels-1)
                    state['step'] = 0
                    state['exp_avg_sq'] = torch.zeros_like(p)

                state['step'] += 1
                bias_correction = 1 - 0.999 ** state['step']
                state['exp_avg_sq'] = 0.999 * state['exp_avg_sq'] + 0.001 * p.grad.data**2
                
                g_normed = p.grad.data / (state['exp_avg_sq']/bias_correction).sqrt()
                g_normed[torch.isnan(g_normed)] = 0
                g_normed.clamp_(-self.g_bound, self.g_bound)
                
                rounded = torch.round(g_normed*lr_factor)*torch.sign(p.data)
                state['integer'] = (state['integer'] + rounded).clamp_(0,levels-1)
                p.data = torch.sign(p.data) * state['max'] * torch.exp(-lr*state['integer'])

        return loss

class Madam(torch.optim.Optimizer):
    r"""Implements the Madam optimisation algorithm.

        ??? info "[Learning compositional functions via multiplicative weight updates](https://arxiv.org/pdf/2006.14560.pdf)"
            **Abstract**: Compositionality is a basic structural feature of both biological and artificial
            neural networks. Learning compositional functions via gradient descent incurs
            well known problems like vanishing and exploding gradients, making careful
            learning rate tuning essential for real-world applications. This paper proves that
            multiplicative weight updates satisfy a descent lemma tailored to compositional
            functions. Based on this lemma, we derive Madam—a multiplicative version of
            the Adam optimiser—and show that it can train state of the art neural network
            architectures without learning rate tuning. We further show that Madam is easily
            adapted to train natively compressed neural networks by representing their weights
            in a logarithmic number system. We conclude by drawing connections between
            multiplicative weight updates and recent findings about synapses in biology

        ??? info "[GitHub @ jxbz/madam](https://github.com/jxbz/madam)"

        Arguments:
            lr (float, required): The learning rate. 0.01 is a good initial value to try.
            p_scale(float, optional): 
            g_bound(float, optional):     
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter], 
        lr: float=0.01, 
        p_scale: float=3.0, 
        g_bound: float=10.0
    ):
        self.p_scale = p_scale
        self.g_bound = g_bound
        defaults = dict(lr=lr)
        super(Madam, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]
                if len(state) == 0:
                    state['max'] = self.p_scale*(p*p).mean().sqrt().item()
                    state['step'] = 0
                    state['exp_avg_sq'] = torch.zeros_like(p)

                state['step'] += 1
                bias_correction = 1 - 0.999 ** state['step']
                state['exp_avg_sq'] = 0.999 * state['exp_avg_sq'] + 0.001 * p.grad.data**2
                
                g_normed = p.grad.data / (state['exp_avg_sq']/bias_correction).sqrt()
                g_normed[torch.isnan(g_normed)] = 0
                g_normed.clamp_(-self.g_bound, self.g_bound)
                
                p.data *= torch.exp( -group['lr']*g_normed*torch.sign(p.data) )
                p.data.clamp_(-state['max'], state['max'])

        return loss


from moai.utils.arguments import ensure_choices
from functools import partial

import torch
import logging

log = logging.getLogger(__name__)

class Classified(torch.nn.Module):
    
    __CHOICES__ = ['eq', 'neq']

    def __init__(self,
        prediction:       str='eq',
        groundtruth:      str='eq',
    ):
        super(Classified, self).__init__()
        ensure_choices(log, 'Classified Prediction Equality', prediction, Classified.__CHOICES__)
        ensure_choices(log, 'Classified Groundtruth Equality', groundtruth, Classified.__CHOICES__)
        self.pred_rel = torch.eq if prediction == 'eq' else torch.ne
        self.gt_rel = torch.eq if groundtruth == 'eq' else torch.ne

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        b, num_classes = pred.shape[:2]
        class_indices = torch.arange(num_classes).to(pred.device)
        classes = torch.argmax(pred, dim=1).view(b, 1, -1)
        gt = gt.view(b, 1, -1)
        correct = self.pred_rel(classes, class_indices) & self.gt_rel(gt, class_indices)
        correct = correct.float()
        if weights is not None:
            correct = correct * weights
        if mask is not None:
            correct = correct[mask]
        return correct.sum(dim=0).squeeze()

TruePositive = partial(Classified, prediction='eq', groundtruth='eq')
TrueNegative = partial(Classified, prediction='neq', groundtruth='neq')
FalsePositive = partial(Classified, prediction='eq', groundtruth='neq')
FalseNegative = partial(Classified, prediction='neq', groundtruth='eq')